#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on May 17, 2021
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy import uic
from qtpy.QtCore import QPoint, QSize, Qt, Slot
from qtpy.QtGui import QStandardItem, QStandardItemModel
from qtpy.QtWidgets import (
    QAbstractItemView, QDialog, QHeaderView, QMenu, QTreeView)

from karabo.common.api import InstanceStatus
from karabogui import icons
from karabogui.singletons.api import get_manager, get_topology

from .utils import get_dialog_ui

HEADER_LABELS = ["Client Id", "Host"]


class ClientTopologyTreeView(QTreeView):
    def __init__(self, parent=None):
        super().__init__(parent)
        model = QStandardItemModel(parent=self)
        model.setHorizontalHeaderLabels(HEADER_LABELS)
        self.setModel(model)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)

        header = self.header()
        header.setSectionResizeMode(QHeaderView.ResizeToContents)
        self.popupWidget = None
        self.setUniformRowHeights(True)

    @Slot()
    def refresh(self):
        model = self.model()
        model.clear()
        model.setHorizontalHeaderLabels(HEADER_LABELS)

        system_hash = get_topology()._system_hash
        if 'client' not in system_hash:
            return

        root_item = model.invisibleRootItem()
        for name, _, attrs in system_hash['client'].iterall():
            status = InstanceStatus(attrs.get('status', 'ok'))
            client_item = QStandardItem(name)
            if status is InstanceStatus.ERROR:
                icon = icons.deviceInstanceError
            else:
                icon = icons.deviceInstance
            client_item.setIcon(icon)
            client_item.setEditable(False)

            host_item = QStandardItem(attrs.get('host', 'None'))
            host_item.setEditable(False)
            root_item.appendRow([client_item, host_item])


class ClientTopologyDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setModal(False)

        ui_file = get_dialog_ui("client_topology.ui")
        uic.loadUi(ui_file, self)

        self.tree_view = ClientTopologyTreeView(self)
        self.ui_refresh.setIcon(icons.refresh)
        self.ui_refresh.clicked.connect(self.tree_view.refresh)

        layout = self.ui_frame
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.tree_view)

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._context_menu)

        # Refresh at start!
        self.tree_view.refresh()

    @Slot(QPoint)
    def _context_menu(self, pos):
        selection_model = self.tree_view.selectionModel()
        index = selection_model.currentIndex()
        if not index.isValid():
            return

        menu = QMenu(parent=self)
        shutdown_action = menu.addAction('Shutdown client')
        shutdown_action.triggered.connect(self.onKillInstance)
        shutdown_action.setIcon(icons.delete)

        menu.exec(self.mapToGlobal(pos))

    @Slot()
    def onKillInstance(self):
        index = self.tree_view.selectionModel().currentIndex()
        deviceId = index.sibling(index.row(), 0).data()
        manager = get_manager()
        manager.shutdownDevice(deviceId, parent=self)

    def sizeHint(self):
        return QSize(640, 480)



from qtpy.QtGui import QIcon

from karabogui.navigation.utils import get_language_icon
from karabogui.topology.system_tree import SystemTreeNode


def test_get_language_icon(gui_app):
    """Ensure  get_language_icon returns the correct icon type."""
    node = SystemTreeNode()
    node.attributes = {"lang": "python"}
    icon = get_language_icon(node)
    assert isinstance(icon, QIcon)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from pyqtgraph import PlotDataItem
from qtpy.QtGui import QBrush, QPen
from traits.api import Constant, Instance

from karabogui.graph.common.api import make_brush, make_pen

from ..base.plot import BasePlot


class ProfilePlot(BasePlot):

    # The y-label of the plot
    y_label = Constant("Intensity")

    # --- plot items ---
    # The line item of the profile plot
    _data_item = Instance(PlotDataItem)

    # The line item of the fit plot
    _fit_item = Instance(PlotDataItem)

    # --- UI attributes ---
    # The default pen and brush of the profile data
    _data_color = 'b'
    _data_pen = Instance(QPen, factory=make_pen(_data_color))
    _data_brush = Instance(QBrush, factory=make_brush(_data_color, alpha=140))

    # The default pen of the fit data
    _fit_color = 'r'
    _fit_pen = Instance(QPen, factory=make_pen(_fit_color))

    def __init__(self, **traits):
        super().__init__(**traits)
        self._data_item = self._add_plot_item()
        self._fit_item = self._add_plot_item()

    # -----------------------------------------------------------------------
    # Public methods

    def set_data(self, x_data, y_data):
        # Assuming that data is discrete, we add the difference to the last
        # value and feed it to the plot. This last is needed to plot all values
        # when stepMode="center" (True), which is then not used.
        if x_data.size <= 1 or y_data.size <= 1:
            return

        offset = x_data[1] - x_data[0]
        x_corrected = np.append(x_data, x_data[-1] + offset)
        self._data_item.setData(x_corrected, y_data,
                                stepMode="center", pen=self._data_pen,
                                brush=self._data_brush, fillLevel=0)

        # Adjust the y range to the data range to not show the pedestal.
        y_range = ("yRange" if self.orientation in ["top", "bottom"]
                   else "xRange")

        data_range = (y_data.min(), y_data.max())
        # Both need to be finite!
        if all(np.isfinite(data_range)):
            self.plotItem.setRange(**{y_range: data_range})

    def set_fit(self, x_data, y_data):
        self._fit_item.setData(x_data, y_data, pen=self._fit_pen)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from .binding_types import (
    BaseBinding, BindingNamespace, BindingRoot, BoolBinding, ByteArrayBinding,
    CharBinding, FloatBinding, HashBinding, ImageBinding, Int8Binding,
    Int16Binding, Int32Binding, Int64Binding, IntBinding, NDArrayBinding,
    NodeBinding, NoneBinding, PipelineOutputBinding, SchemaBinding,
    SignedIntBinding, SlotBinding, StringBinding, TableBinding, Uint8Binding,
    Uint16Binding, Uint32Binding, Uint64Binding, UnsignedIntBinding,
    VectorBinding, VectorBoolBinding, VectorCharBinding, VectorDoubleBinding,
    VectorFloatBinding, VectorHashBinding, VectorInt8Binding,
    VectorInt16Binding, VectorInt32Binding, VectorInt64Binding,
    VectorNoneBinding, VectorNumberBinding, VectorStringBinding,
    VectorUint8Binding, VectorUint16Binding, VectorUint32Binding,
    VectorUint64Binding, WidgetNodeBinding)
from .builder import build_binding
from .config import (
    apply_configuration, apply_default_configuration, apply_fast_data,
    apply_project_configuration, extract_configuration, extract_edits,
    extract_init_configuration, extract_online_edits,
    extract_sparse_configurations, get_config_changes, iterate_binding)
from .enums import (
    NO_CLASS_STATUSES, NO_CONFIG_STATUSES, ONLINE_CONFIG_STATUSES,
    ONLINE_STATUSES, SCHEMA_STATUSES, ProxyStatus)
from .proxy import (
    BaseDeviceProxy, DeviceClassProxy, DeviceProxy, ProjectDeviceProxy,
    PropertyProxy)
from .util import (
    REFERENCE_TYPENUM_TO_DTYPE, attr_fast_deepcopy, get_binding_array_value,
    get_binding_format, get_binding_value, get_dtype_format, get_editor_value,
    get_min_max, get_min_max_size, get_native_min_max, get_numpy_binding,
    has_min_max_attributes, is_signed_vector_integer,
    is_unsigned_vector_integer, is_vector_floating, is_vector_integer,
    realign_hash)
from .validate import (
    convert_string, get_default_value, validate_binding_configuration,
    validate_table_value, validate_value)



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on February 10, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QComboBox
from traits.api import Instance

from karabo.common.scenemodel.api import EditableComboBoxModel
from karabogui.binding.api import BaseBinding, get_editor_value
from karabogui.controllers.api import (
    BaseBindingController, is_proxy_allowed, register_binding_controller)
from karabogui.util import MouseWheelEventBlocker, SignalBlocker


def _is_compatible(binding):
    return len(binding.options) > 0


@register_binding_controller(ui_name="Selection Field", can_edit=True,
                             klassname="EditableComboBox",
                             binding_type=BaseBinding,
                             is_compatible=_is_compatible, priority=20)
class EditableComboBox(BaseBindingController):
    # The scene model class used by this controller
    model = Instance(EditableComboBoxModel, args=())
    # Internal traits
    _filter = Instance(MouseWheelEventBlocker)

    def create_widget(self, parent):
        widget = QComboBox(parent)
        widget.setFrame(False)

        self._filter = MouseWheelEventBlocker(widget)
        widget.installEventFilter(self._filter)
        widget.currentIndexChanged[int].connect(self._on_user_edit)
        widget.setFocusPolicy(Qt.StrongFocus)
        return widget

    def binding_update(self, proxy):
        with SignalBlocker(self.widget):
            self.widget.clear()
            self.widget.addItems([str(o) for o in proxy.binding.options])
            self.widget.setCurrentIndex(-1)

    def value_update(self, proxy):
        value = get_editor_value(proxy)
        if value is None:
            return
        options = proxy.binding.options
        try:
            index = next(i for i, v in enumerate(options)
                         if v == value)
        except StopIteration:
            index = -1
        with SignalBlocker(self.widget):
            self.widget.setCurrentIndex(index)

    def _on_user_edit(self, index):
        if self.proxy.binding is None:
            return
        self.proxy.edit_value = self.proxy.binding.options[index]

    def state_update(self, proxy):
        enable = is_proxy_allowed(proxy)
        self.widget.setEnabled(enable)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on May 22, 2018
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import (
    QItemSelection, QItemSelectionModel, QSortFilterProxyModel, Qt, Signal,
    Slot)

from karabo.common.enums import Interfaces
from karabogui.singletons.api import get_topology


class DeviceFilterModel(QSortFilterProxyModel):
    signalItemChanged = Signal(str, object)  # type, BaseDeviceProxy

    def __init__(self, source_model=None, parent=None):
        super().__init__(parent)
        self._interface = None
        self.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.setFilterRole(Qt.DisplayRole)
        self.setFilterKeyColumn(0)
        self.setRecursiveFilteringEnabled(True)
        self.setDynamicSortFilter(False)
        self.setSourceModel(source_model)
        self.selectionModel = QItemSelectionModel(self, self)
        self.selectionModel.selectionChanged.connect(self.onSelectionChanged)

    def filterAcceptsRow(self, source_row, source_parent):
        model = self.sourceModel()
        source_index = model.index(source_row, self.filterKeyColumn(),
                                   source_parent)
        if source_index.isValid():
            node = source_index.internalPointer()
            if node is None:
                return True
            if self._interface is not None:
                interface = self._check_interface(node.interfaces,
                                                  self._interface)
                if not interface:
                    return False

        return super().filterAcceptsRow(source_row, source_parent)

    def _check_interface(self, mask, bit):
        return (mask & bit) == bit

    def setFilterSelection(self, text):
        """Set a interface bit for the filtering"""
        if text == "All Devices":
            self._interface = None
        else:
            self._interface = Interfaces[text]

    # --------------------------------------------------------------------
    # Index Methods

    def index_ref(self, index):
        source_index = self.mapToSource(index)
        if not source_index.isValid():
            return None
        return source_index.internalPointer()

    def currentIndex(self):
        """Retrieve the current index for context menu actions"""
        return self.selectionModel.currentIndex()

    def indexInfo(self, index):
        source_index = self.mapToSource(index)
        if not source_index.isValid():
            return {}
        return self.sourceModel().indexInfo(source_index)

    def selectIndex(self, index):
        """Select the given `index` of type `QModelIndex` if this is not None

        :param index: Index of the filter model
        """
        if index is None or not index.isValid():
            self.selectionModel.clearSelection()
            return

        self.selectionModel.setCurrentIndex(index,
                                            QItemSelectionModel.ClearAndSelect)

        treeview = super().parent()
        treeview.scrollTo(index)

    @Slot(QItemSelection, QItemSelection)
    def onSelectionChanged(self, selected, deselected):
        source_selection = self.mapSelectionToSource(selected)
        selectedIndexes = source_selection.indexes()
        if not selectedIndexes:
            return

        index = selectedIndexes[0]
        if index is None or not index.isValid():
            return

        node = index.internalPointer()
        if node is None:
            return
        level = node.level

        if level == 0:
            proxy = None
            item_type = 'domain'
        elif level == 1:
            proxy = None
            item_type = 'type'
        elif level == 2:
            deviceId = node.node_id
            proxy = get_topology().get_device(deviceId)
            item_type = 'device'

        self.signalItemChanged.emit(item_type, proxy)



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on August 8, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from enum import Enum

from qtpy.QtCore import QEvent, QRect, QSize, Qt
from qtpy.QtGui import QBrush, QColor, QDoubleValidator, QPalette, QValidator
from qtpy.QtWidgets import (
    QAbstractItemDelegate, QApplication, QComboBox, QDialog, QHBoxLayout,
    QLineEdit, QStyle, QStyledItemDelegate, QStyleOptionButton, QWidget)

from karabo.common.api import State
from karabogui.binding.api import IntBinding, PropertyProxy, VectorHashBinding
from karabogui.controllers.api import get_compatible_controllers
from karabogui.indicators import STATE_COLORS

from .dialog.table_view import TableDialog
from .utils import (
    FIXED_ROW_HEIGHT, ButtonState, handle_default_state, set_fill_rect)

TABLE_BUTTON_TEXT = 'Table Element'
BUTTON_LABEL_PADDING = 5


def _get_table_button_rect(option):
    """Calculate the rectangle of the table edit button based on the given
    `QStyleOptionViewItem` instance
    """
    rect = QRect(option.rect)
    font_metrics = option.fontMetrics
    width = font_metrics.size(Qt.TextSingleLine, TABLE_BUTTON_TEXT).width()
    rect.setWidth(width + BUTTON_LABEL_PADDING * 2)
    return rect


class EditDelegate(QStyledItemDelegate):
    """A QStyledItemDelegate for configurator values
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._button_states = {}
        self._selection_changed = False

    # ----------------------------------------------------------------------
    # Public interface

    def close_editor(self, editor, hint):
        """Workaround for shitty PyQt behavior.
        """
        self.update_model_data(editor.index, hint)

    def current_changed(self, index):
        """The view is telling us that the selection changed
        """
        self._selection_changed = True

    def update_model_data(self, index, hint):
        model = self.parent().model()
        if hint == QAbstractItemDelegate.SubmitModelCache:
            obj = model.index_ref(index)
            # XXX: some extra handling for VectorHashes since they are not
            # editable via an editor
            editable = index.flags() & Qt.ItemIsEditable == Qt.ItemIsEditable
            is_table = isinstance(getattr(obj, 'binding', None),
                                  VectorHashBinding)
            if not self._selection_changed or (editable and is_table):
                model.flush_index_modification(index)
        elif hint == QAbstractItemDelegate.RevertModelCache:
            model.clear_index_modification(index)

    # ----------------------------------------------------------------------
    # Qt interface

    def createEditor(self, parent, option, index):
        """Reimplemented function of QStyledItemDelegate.
        """
        model = index.model()
        obj = model.index_ref(index)
        if isinstance(getattr(obj, 'binding', None), VectorHashBinding):
            return None

        self._selection_changed = False
        return EditWidgetWrapper(obj, index, parent=parent)

    def setEditorData(self, editor, index):
        """Reimplemented function of QStyledItemDelegate.
        """
        if editor.initialized:
            # Don't mess with editors which already received a value!
            # This avoids resetting cursor positions against the user's will.
            return

        model = index.model()
        obj = model.index_ref(index)
        if obj is None:
            return

        # Handle attribute editors specially
        if isinstance(obj, PropertyProxy):
            editor.controller.value_update(obj)
        else:
            editor.controller.value = index.data(Qt.EditRole)
        editor.initialized = True

    def setModelData(self, editor, model, index):
        """Reimplemented function of QStyledItemDelegate.
        """
        column = index.column()
        if column != 2:
            return

        model = index.model()
        obj = model.index_ref(index)
        if obj is None:
            return

        if isinstance(obj, PropertyProxy):
            value = editor.controller.proxy.edit_value
        else:
            value = editor.controller.value
        # Set the data via the model
        model.setData(index, value, Qt.EditRole)

    def sizeHint(self, option, index):
        """Reimplemented function of QStyledItemDelegate.

        XXX: I don't like this, but it's not clear how I would access the
        editor widget for a row (_if_ it exists) to return the correct height
        here.
        """
        return QSize(option.rect.width(), FIXED_ROW_HEIGHT)

    def paint(self, painter, option, index):
        """Reimplemented function of QStyledItemDelegate.
        """
        model = index.model()
        obj = model.index_ref(index)
        if not isinstance(getattr(obj, 'binding', None), VectorHashBinding):
            super().paint(painter, option, index)
            return

        set_fill_rect(painter, option, index)

        self._draw_button(painter, option, index, obj)

    def editorEvent(self, event, model, option, index):
        """Reimplemented function of QStyledItemDelegate.
        """
        handled_types = (QEvent.MouseButtonPress, QEvent.MouseButtonRelease)
        if event.type() in handled_types:
            proxy = model.index_ref(index)
            if isinstance(getattr(proxy, 'binding', None), VectorHashBinding):
                self._handle_event_state(proxy, event, model, option, index)

        return super().editorEvent(
            event, model, option, index)

    # ----------------------------------------------------------------------
    # Private interface

    def _draw_button(self, painter, option, index, proxy):
        """Draw a table button"""
        key = proxy.key
        state = self._button_states.get(key, ButtonState.DISABLED)
        # always allow table button click!
        button_state = handle_default_state(True, state)
        self._button_states[key] = state
        button = QStyleOptionButton()
        if proxy.edit_value is not None:
            palette = QPalette(button.palette)
            color = QColor(*STATE_COLORS[State.CHANGING])
            color.setAlpha(128)
            palette.setBrush(QPalette.Button, QBrush(color))
            button.palette = palette

        button.state = button_state
        button.rect = _get_table_button_rect(option)
        button.text = TABLE_BUTTON_TEXT
        button.features = QStyleOptionButton.AutoDefaultButton
        QApplication.style().drawControl(QStyle.CE_PushButton, button, painter)

    def _handle_event_state(self, proxy, event, model, option, index):
        """Determine the state of a given box's button during user interaction.
        """
        key = proxy.key
        state = self._button_states.get(key, ButtonState.DISABLED)
        allowed = index.flags() & Qt.ItemIsEditable == Qt.ItemIsEditable
        rect = _get_table_button_rect(option)
        if rect.contains(event.pos()):
            if event.type() == QEvent.MouseButtonPress:
                state = ButtonState.PRESSED
            elif state == ButtonState.PRESSED:
                dialog = TableDialog(proxy, allowed, parent=self.parent())
                result = dialog.exec()
                # Only for editable table elements we do actions!
                if allowed:
                    if result == QDialog.Accepted:
                        # XXX: Note that the dialog is passed as an editor
                        # Ensure that the dialog has a member called
                        # `controller` to enable fetching the data of it
                        self.setModelData(dialog, model, index)
                    else:
                        proxy.revert_edit()
        if (state == ButtonState.PRESSED and
                event.type() == QEvent.MouseButtonRelease):
            state = ButtonState.ENABLED
        self._button_states[key] = state
        return state


class EditWidgetWrapper(QWidget):
    """A QWidget container which can be returned as an item's editor."""

    def __init__(self, obj, index, parent=None):
        super().__init__(parent)
        self.setAutoFillBackground(True)

        assert isinstance(obj, PropertyProxy)
        klass = get_compatible_controllers(obj.binding, can_edit=True)[0]
        self.controller = klass(proxy=obj)
        self.controller.create(self)
        # Enable editing for the widget
        self.controller.set_read_only(False)
        self.setFocusProxy(self.controller.widget)

        # Introduce layout to have some border to show
        layout = QHBoxLayout(self)
        layout.setContentsMargins(2, 2, 2, 2)
        layout.addWidget(self.controller.widget)

        # Keep a model index reference for use when the editor is closed by Qt
        self.index = index
        self.initialized = False


# -----------------------------------------------------------------------------
# Attribute editors aren't typical BaseBindingControllers

class EnumAttributeEditor:
    """An editor for attribute values defined by an Enum class."""
    # Subclasses should define this
    enum_klass = Enum

    def __init__(self, binding, parent=None):
        instMsg = "Don't instantiate EnumAttributeEditor directly!"
        assert self.enum_klass is not Enum, instMsg

        self.widget = QComboBox(parent)
        self.widget.setFrame(False)
        self._populate_widget()

    @property
    def value(self):
        return self.widget.itemData(self.widget.currentIndex())

    @value.setter
    def value(self, value):
        try:
            enum_value = self.enum_klass(value)
        except ValueError:
            return
        index = self.widget.findData(enum_value.value)
        if index >= 0:
            self.widget.setCurrentIndex(index)

    def _populate_widget(self):
        def _get_item_text(enum_val):
            # Normalize the ENUM_VALUE_NAME -> Enum value name
            name = enum_val.name.replace('_', ' ').lower().capitalize()
            return f"{enum_val.value} ({name})"

        for e in self.enum_klass:
            text = _get_item_text(e)
            self.widget.addItem(text, e.value)


class IntValidator(QValidator):
    def validate(self, input, pos):
        if input in ('+', '-', ''):
            return self.Intermediate, input, pos

        if not (input.isdigit() or input[0] in '+-' and input[1:].isdigit()):
            return self.Invalid, input, pos

        return self.Acceptable, input, pos


class NumberAttributeEditor:
    """An editor for numerical attribute values
    """

    def __init__(self, binding, parent=None):
        self.widget = QLineEdit(parent)

        if isinstance(binding, IntBinding):
            self.validator = IntValidator(self.widget)
            self._value_cast = int
        else:
            self.validator = QDoubleValidator(self.widget)
            self._value_cast = float

        self.widget.setValidator(self.validator)
        self.widget.textChanged.connect(self._on_text_changed)
        self._normal_palette = self.widget.palette()
        self._error_palette = QPalette(self._normal_palette)
        self._error_palette.setColor(QPalette.Text, Qt.red)

    @property
    def value(self):
        return self._value_cast(self._validate_value())

    @value.setter
    def value(self, value):
        if value is None:
            value = 0
        self.widget.setText(f"{value}")

    def _on_text_changed(self, text):
        self.widget.setPalette(self._normal_palette
                               if self.widget.hasAcceptableInput()
                               else self._error_palette)

    def _validate_value(self):
        if not self.widget.text():
            return 0

        value = self.widget.text()
        state, _, _ = self.validator.validate(value, 0)
        if state == QValidator.Invalid or state == QValidator.Intermediate:
            value = 0
        return value



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import pop_exception_handler, push_exception_handler

from karabo.common.project.api import (
    PROJECT_OBJECT_CATEGORIES, DeviceConfigurationModel, DeviceInstanceModel,
    DeviceServerModel, MacroModel, ProjectModel)
from karabo.common.scenemodel.api import SceneModel
from karabogui.project.controller.build import (
    create_device_server_controller, create_macro_controller,
    create_project_controller, destroy_project_controller)
from karabogui.project.controller.scene import SceneController
from karabogui.singletons.project_model import ProjectViewItemModel


def assert_no_notification_handlers(proj_model):
    """BaseProjectObjectModel sets a handler for all traits. Ignore it."""
    ACCEPTABLE_HANDLERS = ("_anytrait_changed",)

    for name in PROJECT_OBJECT_CATEGORIES:
        t = proj_model.trait(name)
        notifiers = t._notifiers(1)

        acceptable = True
        for n in notifiers:
            if n.handler.__name__ not in ACCEPTABLE_HANDLERS:
                acceptable = False

        if not acceptable:
            names = "\n\t".join([repr(n.handler) for n in notifiers])
            msg = f"Trait {name} still has notifiers!\n\t{names}"
            raise AssertionError(msg)

    # Recurse!
    for sub in proj_model.subprojects:
        assert_no_notification_handlers(sub)


def test_base_controller_project(gui_app, subtests):
    """Test the base project controller"""
    push_exception_handler(lambda *args: None, reraise_exceptions=True)

    macros = [MacroModel()]
    scenes = [SceneModel()]
    dev0 = DeviceInstanceModel(instance_id="dev0")
    dev1 = DeviceInstanceModel(instance_id="dev1")
    servers = [DeviceServerModel(devices=[dev0, dev1])]
    subsubproject = ProjectModel(macros=macros)
    subprojects = [ProjectModel(macros=macros,
                                subprojects=[subsubproject])]
    proj = ProjectModel(macros=macros, scenes=scenes, servers=servers,
                        subprojects=subprojects)

    creators = (create_macro_controller, SceneController,
                create_device_server_controller, create_project_controller)

    insertions, removals = [], []

    def rows_inserted(parent, start, end):
        insertions.append((start, end))

    def rows_removed(parent, start, end):
        removals.append((start, end))

    qt_model = ProjectViewItemModel(parent=None)
    qt_model.rowsInserted.connect(rows_inserted)
    qt_model.rowsRemoved.connect(rows_removed)

    # Cause the controllers to be created and get a ref to the root
    qt_model.root_model = proj
    controller = qt_model.root_controller

    for subgroup, creator in zip(controller.children, creators):
        assert subgroup.child_create is creator
        assert len(subgroup.children) == 1

    proj_servers = proj.servers
    assert len(proj_servers) == len(servers) == 1
    serv = proj_servers[0]
    assert len(serv.devices) == len(servers[0].devices) == 2
    assert serv.devices[0].instance_id == "dev0"
    assert serv.devices[1].instance_id == "dev1"

    assert len(insertions) == 0
    assert len(removals) == 0

    proj.scenes.append(SceneModel())
    assert len(controller.children[1].children) == 2
    assert len(insertions) == 1
    assert insertions[0] == (1, 1)

    subproj = proj.subprojects.pop()
    assert len(controller.children[-1].children) == 0
    assert len(removals) == 1
    assert removals[0] == (0, 0)
    assert_no_notification_handlers(subproj)

    destroy_project_controller(controller)
    assert_no_notification_handlers(proj)

    with subtests.test("Project building children"):
        sc0 = SceneModel()
        conf = DeviceConfigurationModel()
        dev0 = DeviceInstanceModel(instance_id="dev0", configs=[conf])
        dev1 = DeviceInstanceModel(instance_id="dev1", configs=[conf])
        devServ0 = DeviceServerModel(devices=[dev0, dev1])
        subdev0 = DeviceInstanceModel(instance_id="subdev0")
        subdev1 = DeviceInstanceModel(instance_id="subdev1")
        subDevServ0 = DeviceServerModel(devices=[subdev0, subdev1])
        subprojects = [ProjectModel(servers=[subDevServ0])]
        proj = ProjectModel(scenes=[sc0], servers=[devServ0],
                            subprojects=subprojects)

        qt_model = ProjectViewItemModel(parent=None)
        # Cause the controllers to be created and get a ref to the root
        qt_model.root_model = proj
        controller = qt_model.root_controller

        proj_groups = ["macros", "scenes", "servers", "subprojects"]
        assert len(proj_groups) == len(controller.children)

        for child in controller.children:
            if child.trait_name == proj_groups[1]:
                assert len(child.children) == 1
                proj.scenes.pop()
                assert len(child.children) == 0
                proj.scenes.append(SceneModel())
            elif child.trait_name == proj_groups[2]:
                assert len(child.children) == 1
                server_items = child.children
                assert server_items[0].model is devServ0
                device_instances = server_items[0].children
                assert len(device_instances) == 2
                # Add a small patch for the project_devices
                for device_instance in device_instances:
                    device_instance.__dict__["project_device"] = None
                devServ0.devices.pop()
                assert len(server_items[0].children) == 1
                dev2 = DeviceInstanceModel(instance_id="dev2")
                devServ0.devices.append(dev2)
                assert len(server_items[0].children) == 2

    pop_exception_handler()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from .project_device import ProjectDeviceInstance
from .system_topology import SystemTopology
from .system_tree import SystemTree, SystemTreeNode
from .util import (
    get_macro_servers, getTopology, is_device_online, is_macro_online,
    is_server_online)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QPointF, QRectF, QSizeF, Qt, Signal
from qtpy.QtGui import QColor, QPen
from qtpy.QtWidgets import QGraphicsObject


class BaseCanvas(QGraphicsObject):
    """A transparent graphics object for painting items. It is
    instantiated by supplying its geometry (rect) and should be then added to
    the scene (by some addItem()). The paint and mouse interactions should be
    reimplemented. Once the user has finished drawing the
    item using the mouse interactions, the canvas supplies the
    drawn geometry using editingFinished.

    One has to redefine the values:
        self._drawn_geometry -  the emitted geometry (e.g. QRect() for rects)
        self._default_geometry - the emitted geometry when the user cancels
            the operations. (e.g. empty QRect())

    """

    editingFinished = Signal(object)

    def __init__(self, rect):
        super().__init__()
        self._drawn_geometry = None
        self._default_geometry = QPointF()
        self._bounding_rect = rect
        self._pen = QPen(QColor(255, 255, 255))
        self.setZValue(1000)
        self.setCursor(Qt.CrossCursor)

    def paint(self, painter, *args):
        pass

    def mousePressEvent(self, event):
        """When canvas is pressed by left button, the subclass should do
           desired bookeeping. Base class will only repaint.
           When right button is pressed, finished editing by emitting
           default geometry."""
        if event.button() == Qt.LeftButton:
            self.update()
            event.accept()

    def mouseReleaseEvent(self, event):
        """When left button is released, it usually signals the end of the
           dragging motion. The corresponding geometry will be then emitted.
        """
        if event.button() == Qt.LeftButton:
            self.editingFinished.emit(self._drawn_geometry)
            event.accept()
        elif event.button() == Qt.RightButton:
            self.editingFinished.emit(self._default_geometry)
            event.ignore()

    def destroy(self):
        self.editingFinished.disconnect()
        self.unsetCursor()
        self.deleteLater()

    def boundingRect(self):
        return self._bounding_rect


class RectCanvas(BaseCanvas):
    """Canvas for drawing rectangles."""

    def __init__(self, rect):
        super().__init__(rect)
        self._drawn_geometry = QRectF(0, 0, 0, 0)
        self._default_geometry = QRectF()
        self._start_pos = None

    def paint(self, painter, *args):
        """Reimplemented because we want to draw rectangles"""
        painter.setPen(self._pen)
        painter.drawRect(self._drawn_geometry)

    def mousePressEvent(self, event):
        """Reimplemented because we want to record starting position"""
        if event.button() == Qt.LeftButton:
            self._start_pos = event.pos()
            rect = QRectF(self._start_pos, self._start_pos).normalized()
            self._drawn_geometry = rect
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        """Reimplemented because we want to record geometry and draw
           rectangles when dragging"""
        if event.buttons() & Qt.LeftButton:
            pos = event.pos()
            if self.contains(pos):
                rect = QRectF(self._start_pos, pos).normalized()
                self._drawn_geometry = rect
                self.update()
            event.accept()


class PointCanvas(BaseCanvas):

    def __init__(self, rect):
        super().__init__(rect)
        self._drawn_geometry = QRectF(0, 0, 0, 0)
        self._default_geometry = QRectF()

    def mousePressEvent(self, event):
        """Reimplemented because we only care about mouse release"""
        if event.button() == Qt.LeftButton:
            event.accept()
            return
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        """Reimplemented because we want to record the self._drawn_geometry
           first before emitting it in base class."""
        self._drawn_geometry = QRectF(event.pos(), QSizeF(1, 1))
        super().mouseReleaseEvent(event)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.common.project.api import ProjectModel
from karabo.common.project.device import DeviceInstanceModel
from karabo.common.project.device_config import DeviceConfigurationModel
from karabo.common.project.server import DeviceServerModel
from karabo.native import Hash
from karabogui.project.controller.device import DeviceInstanceController
from karabogui.singletons.mediator import Mediator
from karabogui.singletons.project_model import ProjectViewItemModel
from karabogui.testing import singletons

STATUS_ICON = 'karabogui.project.controller.' \
              'device.get_project_device_status_icon'


def test_active_configuration(gui_app):
    with singletons(mediator=Mediator()):
        # Changing configurations might emit signals
        fconfig = Hash()
        fconfig['fkey'] = 'fvalue'
        fconfig['skey'] = 'svalue'
        fconfig = DeviceConfigurationModel(class_id='BazClass',
                                           configuration=fconfig)
        fconfig.initialized = True
        sconfig = Hash()
        sconfig['fkey'] = 'value_diff'
        sconfig = DeviceConfigurationModel(class_id='BazClass',
                                           configuration=sconfig)
        sconfig.initialized = True
        config_list = [fconfig, sconfig]
        model = DeviceInstanceModel(class_id='BazClass',
                                    instance_id='fooDevice',
                                    configs=config_list)
        model.initialized = True
        model.active_config_ref = fconfig.uuid
        DeviceServerModel(server_id='testServer', host='serverFoo',
                          devices=[model])

        def assert_active_configuration():
            assert 'fkey' in fconfig.configuration
            assert fconfig.configuration['fkey'] == 'fvalue'
            assert 'skey' in fconfig.configuration
            assert fconfig.configuration['skey'] == 'svalue'
            assert 'fkey' in sconfig.configuration
            assert sconfig.configuration['fkey'] == 'value_diff'
            assert 'skey' not in sconfig.configuration

        controller = DeviceInstanceController(model=model)
        controller.active_config_changed(sconfig)
        assert_active_configuration()
        controller.active_config_changed(fconfig)
        assert_active_configuration()
        controller.active_config_changed(sconfig)
        assert_active_configuration()


def test_menu(gui_app, mocker):
    _tooltip = "Requires minimum 'OPERATOR' access level"
    proj = ProjectModel(scenes=[], servers=[],
                        subprojects=[])

    qt_model = ProjectViewItemModel(parent=None)
    # Cause the controllers to be created and get a ref to the root
    qt_model.root_model = proj
    controller = qt_model.root_controller

    # Macro group menu-items
    macro_group = controller.children[0]
    assert macro_group.group_name == "Macros"
    mock_project_enabled = mocker.patch(
        "karabogui.project.controller.project_groups.access_role_allowed")
    mock_project_enabled.return_value = False
    menu = macro_group.context_menu(macro_group)

    actions = menu.actions()
    assert not any([act.isEnabled() for act in actions])
    arrange_macros = actions[3]
    assert arrange_macros.text() == "Arrange Macros"
    assert arrange_macros.toolTip() == _tooltip

    mock_project_enabled = mocker.patch(
        "karabogui.project.controller.project_groups.access_role_allowed")
    mock_project_enabled.return_value = True
    menu = macro_group.context_menu(macro_group)
    actions = menu.actions()
    assert all([act.isEnabled() for act in actions])

    # Scene group menu-items
    scene_group = controller.children[1]
    assert scene_group.group_name == "Scenes"

    mock_project_enabled = mocker.patch(
        "karabogui.project.controller.project_groups.access_role_allowed")
    mock_project_enabled.return_value = False
    menu = scene_group.context_menu(scene_group)

    actions = menu.actions()
    arrange_scenes = actions[6]
    assert arrange_scenes.text() == "Arrange Scenes"
    assert arrange_scenes.toolTip() == _tooltip
    assert not arrange_scenes.isEnabled()
    about = actions[-1]
    assert about.text() == "About"
    assert about.isEnabled()

    load_scene = actions[1]
    assert load_scene.text() == "Load scene..."
    assert not load_scene.isEnabled()
    assert load_scene.toolTip() == "Requires minimum 'EXPERT' access level"

    mock_project_enabled = mocker.patch(
        "karabogui.project.controller.project_groups.access_role_allowed")
    mock_project_enabled.return_value = True
    menu = scene_group.context_menu(scene_group)
    actions = menu.actions()
    assert all([act.isEnabled() for act in actions])



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on January 18, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from traits.api import (
    DelegatesTo, HasStrictTraits, Instance, Int, Property, String,
    on_trait_change)

from karabo.common.api import KARABO_SCHEMA_DEFAULT_VALUE, InstanceStatus
from karabo.native import Hash
from karabogui.binding.api import (
    BaseDeviceProxy, DeviceProxy, ProjectDeviceProxy, ProxyStatus,
    apply_default_configuration, apply_project_configuration, extract_edits,
    extract_online_edits)
from karabogui.logger import get_logger
from karabogui.singletons.api import get_topology


class ProjectDeviceInstance(HasStrictTraits):
    """An abstraction of devices needed for projects and workflow design.
    No matter the online/offline state of a device, you can get a
    ``Configuration`` object which is valid and associated with the same
    device id.
    """
    # The current device id. Can be monitored for changes.
    device_id = String
    class_id = String
    server_id = String

    # A weak reference to the system topology device node
    device_node = DelegatesTo('_online_proxy', prefix='topology_node')

    # The proxy for this device
    proxy = Property(Instance(BaseDeviceProxy),
                     depends_on='_online_proxy:online')
    online = DelegatesTo('_online_proxy')
    schema_update = DelegatesTo('_online_proxy')

    # The online/offline proxies and configurations
    _online_proxy = Instance(DeviceProxy)
    _offline_proxy = Instance(ProjectDeviceProxy)

    # Internal storage for applying offline configuration
    _offline_config = Instance(Hash)

    # An event indicates there are changes in offline configurations so
    # the project should prompt for save
    save_project = DelegatesTo('_offline_proxy',
                               prefix='binding.config_update')

    # Monitor count
    _monitor_count = Int(0)

    # The proxy status
    status = Property(depends_on=['_online_proxy:online',
                                  '_offline_proxy:status'])

    # The instance status deliverd by the instance info
    instance_status = Property(
        depends_on=['_online_proxy:topology_node.status'])

    configuration = Property

    def __init__(self, device_id, server_id, class_id):
        super().__init__()
        self._init_object_state(device_id, server_id, class_id)

    def get_user_edited_config_hash(self):
        """Extract user edited values and attributes from the offline proxy
        """
        topology = get_topology()
        # To extract user edits we need a device schema to find default
        # attribute values
        schema = topology.get_schema(self.server_id, self.class_id)

        if schema is not None and len(self._offline_proxy.binding.value) > 0:
            new_config = extract_edits(schema, self._offline_proxy.binding)
            # update our own copy of the device config
            self._offline_config = new_config
            return new_config
        return Hash()

    def get_user_edited_config_hash_online(self):
        """Extract user edited values from the online proxy

        This method requires that the proxy has both an offline and online
        schema.

        returns: success boolean and configuration Hash

                 For more information see `extract_online_edits` of the config
                 module.
        """
        has_offline = len(self._offline_proxy.binding.value) > 0
        has_online = len(self._online_proxy.binding.value) > 0
        if has_offline and has_online:
            topology = get_topology()
            schema = topology.get_schema(self.server_id, self.class_id)
            config = extract_online_edits(
                schema, self._online_proxy.binding)
            return True, config

        return False, Hash()

    def get_class_proxy(self):
        """Retrieve a class schema for this project device instance"""
        topology = get_topology()
        return topology.get_class(self.server_id, self.class_id)

    def get_schema_default_configuration(self):
        """Extract the ``default`` configuration derived from a schema

        The returned configuration Hash consists of all keys which have
        with defaultValue as their value.
        """
        topology = get_topology()
        schema = topology.get_schema(self.server_id, self.class_id)
        config = Hash()
        if schema is not None:
            for key, _, attrs in schema.hash.iterall():
                default = attrs.get(KARABO_SCHEMA_DEFAULT_VALUE)
                if default is not None:
                    config[key] = default

        return config

    def rename(self, device_id='', server_id='', class_id=''):
        """Assign a new device_id, server_id, class_id."""
        device_id = device_id or self.device_id
        server_id = server_id or self.server_id
        class_id = class_id or self.class_id

        # First check to see if anything is changing!
        if (device_id == self.device_id and class_id == self.class_id and
                server_id == self.server_id):
            return

        self._init_object_state(device_id, server_id, class_id)

    def set_project_config_hash(self, config):
        """Forcibly set the offline configuration Hash of the device.
        """
        if config is None:
            return
        # Store the offline config, whenever the device goes from online to
        # offline or its schema changes, we need to apply the config again.
        self._offline_config = config

        # Only apply offline device configuration if the device is offline
        # and we have schema
        if not self.online and len(self._offline_proxy.binding.value) > 0:
            apply_default_configuration(self._offline_proxy.binding)
            # Do not report here...
            apply_project_configuration(config, self._offline_proxy.binding)

    def set_project_config_hash_online(self, config):
        if config is None or not self.online:
            return

        self._offline_config = config
        apply_default_configuration(self._offline_proxy.binding)
        apply_project_configuration(config, self._offline_proxy.binding)

    def start_monitoring(self):
        """Enable monitoring of the online device (when it is online).
        """
        self._monitor_count += 1

    def stop_monitoring(self):
        """Disable monitoring of the online device (when it is online).
        """
        self._monitor_count -= 1

    # ---------------------------------------------------------------------
    # Traits Handlers

    def _get_instance_status(self):
        """Return the online topology information"""
        node = self.device_node
        if node is not None:
            return node.status
        # None status is offline
        return InstanceStatus.NONE

    def _get_configuration(self):
        """Return the configured configuration (`offline`) for this device"""
        return self._offline_config

    def __online_proxy_changed(self, old, new):
        """Keep _monitor_count in sync with _online_proxy."""
        if old is not None and self._monitor_count > 0:
            old.remove_monitor()
        if new is not None and self._monitor_count > 0:
            new.add_monitor()

    def __monitor_count_changed(self, old, new):
        """Handle _monitor_count state transitions"""
        if old == 0 and new == 1:
            self._online_proxy.add_monitor()
        elif old == 1 and new == 0:
            self._online_proxy.remove_monitor()

    def _get_proxy(self):
        if self._online_proxy.online:
            return self._online_proxy
        return self._offline_proxy

    def _get_status(self):
        """Return the proxy status information"""
        if self._online_proxy.online:
            if self.class_id != self._online_proxy.binding.class_id:
                return ProxyStatus.INCOMPATIBLE
            return self._online_proxy.status
        return self._offline_proxy.status

    @on_trait_change('_online_proxy:online,_offline_proxy:schema_update')
    def _apply_offline_config(self):
        """Apply offline device configuration, but only is there is a schema
        """
        if not self.online and len(self._offline_proxy.binding.value) > 0:
            apply_default_configuration(self._offline_proxy.binding)
            device_id = self._offline_proxy.device_id
            if self._offline_config is None:
                text = ("Ignoring corrupted project configuration "
                        "for device {}!".format(device_id))
                get_logger().error(text)
                return

            fails = apply_project_configuration(
                self._offline_config, self._offline_proxy.binding)
            if fails:
                number = len(fails.paths(intermediate=False))
                text = (f"Apply offline configuration for <b>{device_id}</b> "
                        f"reported <b>{number}</b> problem(s)")
                get_logger().error(text)

            # Notify configuration has been applied!
            self._offline_proxy.config_update = True

    @on_trait_change('_offline_proxy:status', post_init=True)
    def _status_changed(self, old, new):
        """Clear the offline proxy binding values when device server goes
        offline
        """
        if new == ProxyStatus.NOSERVER:
            self._offline_proxy.binding.value.clear_namespace()

    # ---------------------------------------------------------------------
    # utils

    def _init_object_state(self, device_id, server_id, class_id):
        """Initialize the object for a given device_id/server_id/class_id"""
        topology = get_topology()

        if self._offline_proxy is not None:
            topology.remove_project_device_proxy(
                self.device_id, self.server_id, self.class_id)

        self._online_proxy = topology.get_device(device_id, request=False)
        self._offline_proxy = topology.get_project_device_proxy(
            device_id, server_id, class_id)

        # Note: Before we were overwriting an empty classId of the online
        # proxy if needed

        # Remember the IDs (also notifies the outside world of changes)
        self.device_id = device_id
        self.class_id = class_id
        self.server_id = server_id



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import re
import weakref

from qtpy import uic
from qtpy.QtCore import Slot
from qtpy.QtGui import QValidator
from qtpy.QtWidgets import QHeaderView, QWidget

import karabogui.icons as icons
from karabo.common.api import State
from karabo.common.enums import Interfaces
from karabogui.indicators import STATE_COLORS
from karabogui.util import wait_cursor

from .utils import get_panel_ui


class SearchValidator(QValidator):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.pattern = re.compile(pattern=r"^[a-zA-Z0-9\-\/\_]+$")

    def validate(self, input, pos):
        if not input:
            return self.Acceptable, input, pos

        if not self.pattern.match(input):
            return self.Invalid, input, pos

        return self.Acceptable, input, pos


_QEDIT_STYLE = """
QLineEdit {{
    border: 1px solid gray;
    border-radius: 2px;
    background-color: rgba{};
}}
QLineEdit:focus
{{
    border: 1px solid rgb(48,140,198);
}}
"""


_QCOMBOX_STYLE = """
QComboBox {{
    border: 1px solid gray;
    border-radius: 2px;
    background-color: rgba{};
    selection-background-color: rgb(48,140,198);
}}
"""

_CHANGED_COLOR = STATE_COLORS[State.CHANGING] + (128,)


class BaseSearchBar(QWidget):
    """The `BaseSearchBar` for filtering a`QAbstractItemModel`

    Note: By default this class does not apply a `QValidator` for text
          filtering.
    """
    ui_file = None

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        uic.loadUi(get_panel_ui(self.ui_file), self)
        self.ui_filter.returnPressed.connect(self._search_clicked)
        self.ui_filter.textChanged.connect(self._filter_modified)
        self.ui_search.clicked.connect(self._search_clicked)
        self.ui_clear.clicked.connect(self._clear_clicked)

        # Set StyleSheets
        self._filter_normal = _QEDIT_STYLE.format((0, 0, 0, 0))
        self._filter_changed = _QEDIT_STYLE.format(_CHANGED_COLOR)
        self.ui_filter.setStyleSheet(self._filter_normal)
        self.tree_view = None
        self._enable_filter()

    def _enable_filter(self, enable=False):
        self.ui_filter.setText("")
        self.ui_filter.setPlaceholderText("")
        self.ui_filter.setEnabled(enable)
        self.ui_search.setEnabled(enable)
        self.ui_clear.setEnabled(enable)

    def _set_filter_modified(self, modified):
        sheet = self._filter_changed if modified else self._filter_normal
        self.ui_filter.setStyleSheet(sheet)

    # -----------------------------------------
    # Public interface

    def setView(self, view):
        self.tree_view = weakref.ref(view)

    def reset(self, enable=False):
        """Reset the model search bar"""
        self._enable_filter(enable)
        self._clear_clicked()

    # -----------------------------------------
    # Qt Slots

    @Slot()
    def _filter_modified(self):
        self._set_filter_modified(True)

    @Slot()
    def _search_clicked(self):
        with wait_cursor():
            self.on_search()
            pattern = str(self.ui_filter.text())
            self.ui_filter.setPlaceholderText(pattern)

            # View and activate
            tree_view = self.tree_view()
            proxy_model = tree_view.model()
            proxy_model.setFilterFixedString(pattern)
            tree_view.expandAll()
            # Expand whole tree and scroll try to keep the selection
            index = self.currentIndex()
            if index.isValid():
                tree_view.scrollTo(index)
            self._set_filter_modified(False)
            self.post_search()

    @Slot()
    def _clear_clicked(self):
        with wait_cursor():
            self.on_clear()
            pattern = ''
            self.ui_filter.setText(pattern)
            self.ui_filter.setPlaceholderText(pattern)

            # View and activate
            tree_view = self.tree_view()
            proxy_model = tree_view.model()
            proxy_model.setFilterFixedString(pattern)
            # Collapse whole tree and scroll / expand the selection
            tree_view.collapseAll()
            index = self.currentIndex()
            if index.isValid():
                tree_view.scrollTo(index)
            self._set_filter_modified(False)
            self.post_clear()

    # -----------------------------------------
    # Subclass functions

    def on_search(self):
        """Subclass this method to act on searching of a pattern"""

    def on_clear(self):
        """Subclass this method to act on clearing of a pattern"""

    def post_search(self):
        """Subclass this method to act after searching of a pattern"""

    def post_clear(self):
        """Subclass this method to act after clearing of a pattern"""

    def currentIndex(self):
        """Subclass this method to retrieve the current index"""
        view = self.tree_view()
        return view.model().currentIndex()


class BaseSelectionBar(BaseSearchBar):
    """The `BaseSelectionBar` class that is used in the topology filtering

    This class applies a `QValidator` for instance filtering
    """
    ui_file = "combo_filter.ui"

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.ui_filter.setValidator(SearchValidator())
        self.setup_combo()
        # Set StyleSheets
        self._combo_normal = _QCOMBOX_STYLE.format((0, 0, 0, 0))
        self._combo_changed = _QCOMBOX_STYLE.format(_CHANGED_COLOR)
        self.ui_combo.setStyleSheet(self._combo_normal)
        self.ui_combo.currentIndexChanged.connect(self._combo_modified)

    def setup_combo(self):
        """Subclass this method to setup the combo filter box"""

    # -----------------------------------------

    def on_search(self):
        model = self.tree_view().model()
        model.setFilterSelection(self.ui_combo.currentText())

    def on_clear(self):
        self.ui_combo.setCurrentIndex(0)
        model = self.tree_view().model()
        model.setFilterSelection(self.ui_combo.currentText())

    def post_search(self):
        self._set_combo_modified(False)

    def post_clear(self):
        self._set_combo_modified(False)

    # -----------------------------------------
    # Qt Slots

    @Slot()
    def _combo_modified(self):
        self._set_combo_modified(True)

    def _set_combo_modified(self, modified):
        sheet = self._combo_changed if modified else self._combo_normal
        self.ui_combo.setStyleSheet(sheet)


class SearchBar(BaseSelectionBar):

    def setup_combo(self):
        self.ui_combo.addItem("No Status Filtering")
        self.ui_combo.addItem("Health Status [OK]")
        self.ui_combo.addItem("Health Status [ERROR]")
        self.ui_combo.addItem("Health Status [UNKNOWN]")
        self.ui_combo.setItemIcon(1, icons.statusOk)
        self.ui_combo.setItemIcon(2, icons.statusError)
        self.ui_combo.setItemIcon(3, icons.statusUnknown)


class InterfaceBar(BaseSelectionBar):

    def setup_combo(self):
        self.ui_combo.addItem("All Devices")
        for interface in Interfaces:
            self.ui_combo.addItem(interface.name)


class ConfiguratorSearch(BaseSearchBar):
    ui_file = "configurator_filter.ui"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.active = False
        self.setVisible(False)

    def currentIndex(self):
        return self.tree_view().currentIndex()

    @Slot()
    def toggleActivate(self):
        """Toggle the activation and swap of the qt models"""
        self.tree_view().swap_models()
        active = not self.active
        if active:
            self.reset(True)
        self.setVisible(active)
        # Finally, resize when phasing out!
        if not active:
            self.resize_contents()
        self.active = active

    def resize_contents(self):
        header = self.tree_view().header()
        header.resizeSections(QHeaderView.ResizeToContents)

    # -----------------------------------------

    def on_search(self):
        self.tree_view().close_popup_widget()

    def on_clear(self):
        self.tree_view().close_popup_widget()

    def post_search(self):
        self.resize_contents()

    def post_clear(self):
        self.resize_contents()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on July 11, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from abc import abstractmethod
from functools import partial

from qtpy.QtCore import QPoint, QRect
from qtpy.QtWidgets import QAction, QDialog, QMenu
from traits.api import ABCHasStrictTraits, Any

from karabo.common.scenemodel.api import BaseWidgetObjectData
from karabogui import icons
from karabogui.binding.proxy import ONLINE_STATUSES
from karabogui.controllers.api import (
    get_class_const_trait, get_compatible_controllers, get_scene_model_class)
from karabogui.dialogs.api import ProxiesDialog, SceneItemDialog
from karabogui.sceneview.tools.api import send_to_back, send_to_front, ungroup
from karabogui.sceneview.widget.api import ControllerContainer


class SceneWidgetHandler(ABCHasStrictTraits):
    # Current widget
    widget = Any

    def handle(self, scene_view, event):
        """Handle the widget and event."""
        menu = QMenu(self.widget)
        move_action = QAction(icons.move, "Move Layout", self.widget)
        move_action.triggered.connect(partial(self._move_dialog, scene_view))
        menu.addAction(move_action)

        # NOTE: Only if we don't have more items selected than a single one, we
        # are allowed to resize the layout!
        if not len(scene_view.selection_model) > 1:
            resize_action = QAction(icons.resize, "Resize Layout", self.widget)
            resize_action.triggered.connect(partial(self._resize_dialog,
                                                    scene_view))
            menu.addAction(resize_action)
        ungroup_action = QAction(icons.ungroup, "Ungroup Layout", self.widget)
        ungroup_action.triggered.connect(partial(self._ungroup,
                                                 scene_view))
        menu.addAction(ungroup_action)

        to_front_action = QAction(icons.bringToFront, "Send To Front",
                                  self.widget)
        to_front_action.triggered.connect(partial(self._to_front,
                                                  scene_view))
        menu.addAction(to_front_action)
        to_back_action = QAction(icons.sendToBack, "Send To Back", self.widget)
        to_back_action.triggered.connect(partial(self._to_back,
                                                 scene_view))
        menu.addAction(to_back_action)

        self.handle_widget(scene_view, event, menu)

    # ------------------------------
    # Private interface

    def _move_dialog(self, scene_view):
        """Move the layout selection on the scene view via a dialog interaction
        """
        selection_model = scene_view.selection_model
        rect = selection_model.get_item_rect()
        max_x = scene_view.scene_model.width - rect.width()
        max_y = scene_view.scene_model.height - rect.height()
        dialog = SceneItemDialog(x=rect.x(), y=rect.y(),
                                 title='Move Layout',
                                 max_x=max_x, max_y=max_y, parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            x_coord = dialog.x - rect.x()
            y_coord = dialog.y - rect.y()
            offset = QPoint(x_coord, y_coord)
            for c in selection_model:
                c.translate(offset)
            scene_view.update()

    def _resize_dialog(self, scene_view):
        """Resize the given layout selection in the scene view"""
        selection_model = scene_view.selection_model
        rect = selection_model.get_item_rect()
        max_x = scene_view.scene_model.width - rect.x()
        max_y = scene_view.scene_model.height - rect.y()
        dialog = SceneItemDialog(x=rect.width(), y=rect.height(),
                                 title='Resize Layout',
                                 max_x=max_x, max_y=max_y, parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            new_rect = QRect(rect.x(), rect.y(),
                             dialog.x, dialog.y)
            for c in selection_model:
                c.set_geometry(new_rect)
            scene_view.update()

    def _to_back(self, scene_view):
        send_to_back(scene_view)

    def _to_front(self, scene_view):
        send_to_front(scene_view)

    def _ungroup(self, scene_view):
        ungroup(scene_view)

    # ------------------------------
    # Abstract public interface

    @abstractmethod
    def can_handle(self):
        """Check whether the event can be handled."""

    @abstractmethod
    def handle_widget(self, scene_view, event, menu):
        """Handle the event from the sceneview and attach actions to the menu!
        """


class SceneToolHandler(SceneWidgetHandler):
    """SceneTool handler to provide operations on workflow items and layouts
    """

    def can_handle(self):
        return not isinstance(self.widget, ControllerContainer)

    def handle_widget(self, scene_view, event, menu):
        # The non-controller widget might have some actions for us!
        if self.widget.actions():
            menu.addSeparator()
            menu.addActions(self.widget.actions())
        menu.exec(event.globalPos())


class SceneControllerHandler(SceneWidgetHandler):
    """Scene Controller handler to provide basic operation on our SceneView

    The widget handler can be used to change the controller class
    """

    def can_handle(self):
        return isinstance(self.widget, ControllerContainer)

    def handle_widget(self, scene_view, event, menu):
        """A callback which is fired whenever the user requests a context menu
        in the SceneView.
        """
        controller = self.widget.widget_controller

        # We don't allow changing widgets without binding!
        binding = controller.proxy.binding
        status = controller.proxy.root_proxy.status
        if binding is None or status not in ONLINE_STATUSES:
            info = menu.addAction('No mutation for offline properties')
            info.setEnabled(False)
            menu.exec(event.globalPos())
            return

        # Add actions which are bound to the actual Qt widget
        qwidget = controller.widget
        if qwidget.actions():
            name = get_class_const_trait(controller, '_ui_name')
            property_menu = menu.addMenu(icons.undefinedAttribute,
                                         f'{name}: Properties')
            property_menu.addActions(qwidget.actions())
            property_menu.addSeparator()

        # But we don't allow to mutate a controller which contains many proxies
        if len(controller.proxies) > 1:
            remove_action = menu.addAction('Remove additional property')
            remove_action.triggered.connect(self._change_widget_proxies)
            menu.exec(event.globalPos())
            return

        # We can finally mutate our widget!
        model = controller.model
        if model.parent_component == 'EditableApplyLaterComponent':
            can_edit = True
        else:
            can_edit = False

        mutate_menu = menu.addMenu(icons.change, 'Change Widget')
        klasses = get_compatible_controllers(binding, can_edit=can_edit)
        klasses.sort(key=lambda w: get_class_const_trait(w, '_ui_name'))
        for klass in klasses:
            ui_name = get_class_const_trait(klass, '_ui_name')
            action = mutate_menu.addAction(ui_name)
            action.triggered.connect(partial(self._change_widget,
                                             scene_view, klass))
            if isinstance(controller, klass):
                action.setCheckable(True)
                action.setChecked(True)

        menu.exec(event.globalPos())

    # -------------------------------------------------
    # Private interface

    def _change_widget(self, scene_view, klass):
        """The model of ``self.widget`` is about to be changed.

        The given ``klass`` object is a `BaseBindingController` class which the
        widget is about to be replaced with.
        """
        model_klass = get_scene_model_class(klass)

        old_model = self.widget.model
        # Get old model traits
        traits = {}
        for key in BaseWidgetObjectData.class_editable_traits():
            traits[key] = getattr(old_model, key)
        # Set the `klass`, if settable
        if 'klass' in model_klass.class_editable_traits():
            traits['klass'] = get_class_const_trait(klass, '_klassname')
        # Ignore width and height for correct layout recalculation
        traits['height'] = traits['width'] = 0
        new_model = model_klass(**traits)
        klass.initialize_model(self.widget.widget_controller.proxy, new_model)
        scene_view.replace_model(old_model, new_model)

    def _change_widget_proxies(self):
        ProxiesDialog(self.widget, parent=self.widget).open()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on November 7, 2011
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import socket
from os import environ, getpid, path
from platform import system

from karabo.common.packaging import utils
from karabo.common.scenemodel.api import SCENE_DEFAULT_DPI, SCENE_MAC_DPI

# APPLICATION
APPLICATION_MODE = False

# WIDGET PROPERTIES
WIDGET_MIN_WIDTH = 40
WIDGET_MIN_HEIGHT = 18

# SERVER COMMUNICATION PROPERTIES
REQUEST_REPLY_TIMEOUT = 5  # in seconds

# Presentation
GUI_DPI_FACTOR = (
    SCENE_DEFAULT_DPI / SCENE_MAC_DPI if system() == "Darwin" else 1)

# Operating System relevant
OS_SYSTEM = system()
IS_MAC_SYSTEM = OS_SYSTEM == "Darwin"
IS_WINDOWS_SYSTEM = OS_SYSTEM == "Windows"
IS_LINUX_SYSTEM = OS_SYSTEM == "Linux"

# Hidden karabo folder which includes certain karabo related files
if IS_WINDOWS_SYSTEM:
    HIDDEN_KARABO_FOLDER = path.join(environ['APPDATA'], 'karabo')
else:
    HIDDEN_KARABO_FOLDER = path.join(environ['HOME'], '.karabo')

# Project folder
KARABO_PROJECT_FOLDER = path.join(HIDDEN_KARABO_FOLDER, 'projects')
# Karabo GUI Client ID
KARABO_CLIENT_ID = f"{socket.gethostname()}-{getpid()}"
CLIENT_HOST = socket.gethostname()

# Tooltips
TOOLTIP_RUNTIME_CONFIG = (
    '<table width="300"><tr><td>'
    'A runtime configuration can be retrieved from the Datalogger System '
    'and may include read-only or dynamic properties. The lookup is '
    'based on a timestamp.</td></tr></table>'
)

TOOLTIP_INIT_CONFIG = (
    '<table width="300"><tr><td>'
    'An init configuration can be retrieved from the Configuration Manager '
    'and includes only writable properties set at instantiation that differ '
    'from the device\'s default settings. The lookup is performed '
    'using a name.</td></tr></table>'
)

try:
    from karabogui import _version

    version = _version.version
    if hasattr(_version, 'full_version'):
        full_version = getattr(_version, 'full_version')
    else:
        full_version = version

    GUI_VERSION = utils.extract_base_version(version)
    GUI_VERSION_LONG = utils.extract_full_version(version)
    GUI_VERSION_DETAILED = version
except (ImportError, AttributeError):
    print('Version file not found! Please generate the _version.py file.')
    exit(1)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from .aux_plots.base.plot import SHOWN_AXES, AuxPlotItem
from .aux_plots.controller import AuxPlotsController
from .aux_plots.histogram.analyzer import HistogramAnalyzer
from .aux_plots.histogram.controller import (
    HistogramAggregator, HistogramAnalyzer)
from .aux_plots.histogram.plot import HistogramPlot
from .aux_plots.histogram.stats import HistogramStats
from .aux_plots.profiling.analyzer import ProfileAnalyzer
from .aux_plots.profiling.controller import (
    ProfileAggregator, ProfileController)
from .aux_plots.profiling.plot import ProfilePlot
from .aux_plots.profiling.stats import ProfileStats
from .base import KaraboImageView
from .colorbar import ColorBarWidget, ColorViewBox
from .dialogs.levels import LevelsDialog
from .image_node import KaraboImageNode
from .item import KaraboImageItem, karabo_default_image
from .legends.picker import PickerLegend
from .legends.scale import ScaleLegend
from .plot import KaraboImagePlot
from .roll_image import RollImage
from .tools.picker import PickerController
from .tools.toolbar import AuxPlotsToolset
from .utils import (
    beam_profile_table_html, bytescale, create_colormap_menu,
    create_icon_from_colormap, ensure_finite_levels, get_transformation,
    levels_almost_equal, rescale)
from .viewbox import KaraboImageViewBox



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import Constant, Dict

from karabogui.graph.common.formatting import (
    table_body, table_header, table_row)

from ..base.stats import BaseStats


class ProfileStats(BaseStats):

    # This contains statistics for both x- and y-axis.
    x_stats = Dict
    y_stats = Dict

    # HTML table format for the stats
    html_table = Constant(
        "<b><u>Beam Profile Statistics</b></u><br/>"
        + table_row(table_header(tabs=('x', 'y')))
        + table_row(table_body(header="Amplitude",
                               tabs=("{x_ampl}", "{y_ampl}")))
        + table_row(table_body(header="Max Pos.",
                               tabs=("{x_maxpos}", "{y_maxpos}")))
        + table_row(table_body(header="FWHM",
                               tabs=("{x_fwhm}", "{y_fwhm}")))
    )

    # -----------------------------------------------------------------------
    # Trait properties

    def _get_html(self):
        html = ''
        x_stats, y_stats = self.x_stats, self.y_stats
        if x_stats and y_stats:
            table = self.html_table.format(
                x_ampl=self._to_string(x_stats.get("amplitude")),
                x_maxpos=self._to_string(x_stats.get("max_pos")),
                x_fwhm=self._to_string(x_stats.get("fwhm")),
                y_ampl=self._to_string(y_stats.get("amplitude")),
                y_maxpos=self._to_string(y_stats.get("max_pos")),
                y_fwhm=self._to_string(y_stats.get("fwhm")))
            html = self.html_format.format(contents=table)
        return html



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtWidgets import QDialog

from karabo.common.scenemodel.api import DisplayTimeModel
from karabo.native import Configurable, Float, Hash, Timestamp
from karabogui.fonts import get_font_size_from_dpi
from karabogui.testing import get_class_property_proxy, set_proxy_hash

from ..timelabel import DisplayTimeLabel


class Object(Configurable):
    prop = Float()


@pytest.fixture
def time_label_setup(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = DisplayTimeLabel(proxy=proxy, model=DisplayTimeModel())
    controller.create(None)
    yield controller, proxy


def test_time_label_basics(time_label_setup):
    controller, _ = time_label_setup
    assert controller.widget is not None

    controller.destroy()
    assert controller.widget is None

    model = controller.model
    assert model.time_format == "%H:%M:%S"
    assert model.font_size == 10
    assert model.font_weight == "normal"


def test_time_label_set_value(time_label_setup):
    controller, proxy = time_label_setup
    property_hash = Hash("prop", 2.0)
    time_stamp = Timestamp("2009-04-20T10:32:22")
    set_proxy_hash(proxy, property_hash, time_stamp)
    assert controller.widget.text() == "10:32:22"


def test_change_time_format(time_label_setup, mocker):
    controller, proxy = time_label_setup
    property_hash = Hash("prop", 4.0)
    time_stamp = Timestamp("2012-04-20T10:35:27")
    set_proxy_hash(proxy, property_hash, time_stamp)
    action = controller.widget.actions()[0]
    assert action.text() == "Change datetime format..."

    dsym = "karabogui.controllers.display.timelabel.QInputDialog"
    QInputDialog = mocker.patch(dsym)
    QInputDialog.getText.return_value = "%H:%M", True
    action.trigger()
    assert controller.widget.text() == "10:35"


def test_format_field(time_label_setup, mocker):
    """
    Test contex menu-item opens the dialog and the model values are
    set from the dialog.
    """
    controller, _ = time_label_setup
    action = controller.widget.actions()[1]
    assert action.text() == "Format field..."

    path = "karabogui.controllers.display.timelabel.FormatLabelDialog"
    dialog = mocker.patch(path)
    dialog().exec.return_value = QDialog.Accepted
    dialog().font_size = 11
    dialog().font_weight = "bold"
    action.trigger()
    assert dialog().exec.call_count == 1
    assert controller.model.font_size == 11
    assert controller.model.font_weight == "bold"


def test_apply_format(time_label_setup):
    """
    Test '_apply_format' applies the font size and weight to widget from
    the model.
    """
    controller, _ = time_label_setup
    widget = controller.widget
    controller._apply_format(widget)
    font = widget.font()
    size = font.pointSize()
    bold = font.bold()

    assert size == get_font_size_from_dpi(controller.model.font_size)
    assert bold == (controller.model.font_weight == "bold")



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 16, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from pathlib import Path

from qtpy.QtCore import QByteArray
from qtpy.QtSvg import QSvgRenderer

from karabo.common.scenemodel.api import (
    LabelModel, LineModel, RectangleModel, UnknownWidgetDataModel)
from karabogui import icons
from karabogui.sceneview.api import (
    LabelWidget, LineShape, RectangleShape, SceneSelectionModel,
    UnknownSvgWidget, UnknownWidget)


def test_selection_model_types(gui_app):
    selection_model = SceneSelectionModel()
    assert not selection_model.has_selection()
    endpoint_x, endpoint_y = (100, 100)

    line_model = LineModel(x1=0, y1=0, x2=endpoint_x, y2=endpoint_y)
    line_shape = LineShape(model=line_model)
    selection_model.select_object(line_shape)
    assert selection_model.has_selection()

    selection_model.clear_selection()
    assert not selection_model.has_selection()

    rect_model = RectangleModel(x=endpoint_x, y=endpoint_y,
                                width=100, height=100)
    rect_shape = RectangleShape(model=rect_model)
    selection_model.select_object(rect_shape)
    assert selection_model.has_selection()

    selection_model.clear_selection()
    assert not selection_model.has_selection()

    selection_model.select_object(line_shape)
    selection_model.select_object(rect_shape)

    for obj in selection_model:
        assert obj in (line_shape, rect_shape)

    exp_rect = line_shape.geometry().united(rect_shape.geometry())
    sel_rect = selection_model.get_selection_bounds()
    assert exp_rect == sel_rect

    selection_model.deselect_object(line_shape)
    assert selection_model.has_selection()

    selection_model.deselect_object(rect_shape)
    assert not selection_model.has_selection()


def test_selection_model_unknown(gui_app):
    selection_model = SceneSelectionModel()
    assert not selection_model.has_selection()

    label_model = LabelModel()
    label_widget = LabelWidget(label_model)

    unknown_model = UnknownWidgetDataModel(
        klass="Notthere",
        data="",
        attributes={})
    unknown_widget = UnknownWidget(unknown_model)

    selection_model.select_object(label_widget)
    selection_model.select_object(unknown_widget)
    assert selection_model.has_selection()
    assert len(selection_model) == 2

    path = Path(icons.__file__).parent / "about.svg"
    with open(path, "rb") as fp:
        b = fp.read()

    ar = QByteArray.fromRawData(b)
    renderer = QSvgRenderer(ar)
    assert renderer.isValid()
    svg_widget = UnknownSvgWidget(renderer=renderer)

    # Try to select an unknown svg widget does nothing
    selection_model.select_object(svg_widget)
    assert selection_model.has_selection()
    assert len(selection_model) == 2



from qtpy import uic
from qtpy.QtCore import QObject, Qt, Slot
from qtpy.QtGui import QPalette
from qtpy.QtWidgets import QDialog
from traits.api import Undefined

from karabo.native import AccessMode, Assignment, Hash, create_html_hash
from karabogui.binding.api import DeviceClassProxy, iterate_binding
from karabogui.dialogs.utils import get_dialog_ui
from karabogui.singletons.api import get_topology
from karabogui.util import get_spin_widget

CONFIGURATION_PAGE = 0
WAITING_PAGE = 1
INIT_CONFIG = "Init Configuration"
CONFIG_WITH_DEFAULTS = "Init Configuration with Schema defaults"


def extract_defaults(binding) -> Hash:
    """Extract the default values from the class proxy binding."""
    ret = Hash()
    for key, node in iterate_binding(binding):
        read_only = node.accessMode is AccessMode.READONLY
        is_internal = node.assignment is Assignment.INTERNAL
        if read_only or is_internal:
            continue
        default = node.value
        if default is Undefined:
            continue
        ret[key] = default

    return ret


class InitConfigurationPreviewDialog(QDialog):
    def __init__(self, proxy: DeviceClassProxy, configuration: Hash,
                 title: str, parent: QObject | None = None):

        super().__init__(parent)

        ui_file = get_dialog_ui("init_configuration_preview.ui")
        uic.loadUi(ui_file, self)
        self.setWindowTitle(title)
        self.configuration = configuration
        self._config_with_defaults = None

        wait_widget = get_spin_widget(icon='wait', parent=self)
        wait_widget.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        wait_widget.setAutoFillBackground(True)
        wait_widget.setBackgroundRole(QPalette.Base)
        self.stackedWidget.addWidget(wait_widget)

        self.tabWidget.setTabText(0, INIT_CONFIG)
        self.tabWidget.setTabText(1, CONFIG_WITH_DEFAULTS)

        self.ui_configuration_selector.addItems(
            [INIT_CONFIG, CONFIG_WITH_DEFAULTS])
        self.ui_configuration_selector.currentIndexChanged.connect(
            self._update_info_label)

        server_id = proxy.server_id
        class_id = proxy.binding.class_id

        class_proxy = get_topology().get_class(server_id, class_id)
        if not len(class_proxy.binding.value) > 0:

            def config_handler():
                class_proxy.on_trait_change(
                    config_handler, "config_update", remove=True)
                self.show_configuration(class_proxy.binding)

            class_proxy.on_trait_change(config_handler, "config_update")

            # Schema has not arrived yet. Let's show the waiting page.
            self.stackedWidget.setCurrentIndex(WAITING_PAGE)
        else:
            self.show_configuration(class_proxy.binding)

        self._update_info_label(index=0)

    @property
    def configuration_to_apply(self) -> Hash:
        if self.ui_configuration_selector.currentText() == INIT_CONFIG:
            return self.configuration
        return self._config_with_defaults

    def show_configuration(self, binding):
        default_values = extract_defaults(binding)
        default_values.merge(self.configuration)
        self._config_with_defaults = default_values
        self.ui_configuration_text.setHtml(create_html_hash(
            self.configuration))
        self.ui_full_configuration_text.setHtml(create_html_hash(
            default_values))
        self.stackedWidget.setCurrentIndex(CONFIGURATION_PAGE)

    @Slot(int)
    def _update_info_label(self, index):
        if index == 0:
            text = ("Only the init configuration is applied to the "
                    "Configuration Editor.")
        else:
            text = ("The init configuration with class schema defaults "
                    "is applied to the Configuration Editor.")
        self.ui_info_label.setText(text)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np

from karabogui.graph.common.api import ImageRegion
from karabogui.testing import GuiTestCase

from ..controller import (
    HistogramAggregator, HistogramAnalyzer, HistogramController, HistogramPlot)


class TestHistogramAggregator(GuiTestCase):

    def setUp(self):
        super().setUp()
        self._aggregator = HistogramAggregator()
        self._aggregator.on_trait_change(self._mock_slot, "stats")

    def tearDown(self):
        super().tearDown()
        self._aggregator.on_trait_change(self._mock_slot, "stats", remove=True)
        self._emitted_value = None

    def _mock_slot(self, value):
        self._emitted_value = value

    def test_basics(self):
        assert len(self._aggregator.controllers) == 1
        aux_plot = self._aggregator.controller
        assert isinstance(aux_plot, HistogramController)
        assert isinstance(aux_plot.plot, HistogramPlot)
        assert isinstance(aux_plot.analyzer, HistogramAnalyzer)
        assert self._aggregator.show_stats

    def test_analyze(self):
        curve_item = self._aggregator.controller.plot._data_item
        analyzer = self._aggregator.controller.analyzer

        # 1. Check default
        np.testing.assert_array_equal(analyzer._hist, [])
        np.testing.assert_array_equal(analyzer._edges, [])
        np.testing.assert_array_equal(curve_item.xData, [])
        np.testing.assert_array_equal(curve_item.yData, [])

        # 2. Check valid region
        self._analyze_region(valid=True)
        assert analyzer._hist is not None
        assert analyzer._edges is not None
        np.testing.assert_array_equal(curve_item.xData, analyzer._edges)
        np.testing.assert_array_equal(curve_item.yData, analyzer._hist)

        # 3. Check invalid region
        self._analyze_region(valid=False)
        np.testing.assert_array_equal(analyzer._hist, [])
        np.testing.assert_array_equal(analyzer._edges, [])
        np.testing.assert_array_equal(curve_item.xData, [])
        np.testing.assert_array_equal(curve_item.yData, [])

    def test_show_stats(self):
        # 1. Check default
        self._assert_show_stats(valid=False)

        # 2. Check valid region
        self._analyze_region(valid=True)
        self._assert_show_stats(valid=True)

        # 3. Check invalid region
        self._analyze_region(valid=False)
        self._assert_show_stats(valid=False)

    def _analyze_region(self, valid=True):
        region = ImageRegion()
        if valid:
            region = ImageRegion(region=np.arange(20).reshape(4, 5),
                                 region_type=ImageRegion.Area,
                                 x_slice=slice(5),
                                 y_slice=slice(4))

        assert region.valid() is valid
        self._aggregator.process(region)

    def _assert_show_stats(self, valid=True):
        assertion = self.assertNotEqual if valid else self.assertEqual

        # Start from disabled stats
        if self._aggregator.show_stats:
            self._aggregator.show_stats = False
            assert self._emitted_value is None

        # Show stats
        self._aggregator.show_stats = True
        assertion(self._emitted_value.html, '')

        # Hide stats
        self._aggregator.show_stats = False
        assert self._emitted_value is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path

import karabogui.icons as icons
from karabogui.const import HIDDEN_KARABO_FOLDER

LOGO_PATH = str(Path(icons.__file__).parent / "splash.png")
LOGO_WIDTH = 50

ICONS_FOLDER = Path(HIDDEN_KARABO_FOLDER) / "icons"
DEFAULT_ICON_PATH = ICONS_FOLDER / Path(LOGO_PATH).name



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os.path as op
from contextlib import contextmanager
from glob import glob

import numpy as np
from traits.api import Undefined

from karabo.common.api import State
from karabo.native import (
    AccessLevel, AccessMode, Assignment, Configurable, Hash, HashList, Int32,
    Schema, String, Timestamp, VectorHash, decodeBinary)
from karabogui.testing import (
    ALL_PROPERTIES_MAP, get_all_props_schema, get_simple_props_schema,
    get_simple_schema)

from ..api import (
    apply_configuration, apply_default_configuration, apply_fast_data,
    apply_project_configuration, build_binding, extract_configuration,
    extract_edits, extract_init_configuration, extract_online_edits)

TEST_DATA_DIR = op.join(op.dirname(__file__), 'data')


def _dict_diff(d0, d1):
    """Get the new items in dictionary d1"""
    return {k: v for k, v in d1.items() if k not in d0 or d0[k] != v}


@contextmanager
def watch_config_update_notification(binding, expected):
    """Watch for binding.config_update notification, assert that trait event
    should be fired/not fired according to `expected`
    """
    notified = False

    def _event_fired():
        nonlocal notified
        notified = True

    try:
        binding.on_trait_event(_event_fired, 'config_update')
        yield
    finally:
        binding.on_trait_event(_event_fired, 'config_update', remove=True)
        assert notified == expected


def test_data_files():
    for filename in glob(op.join(TEST_DATA_DIR, '*.schema')):
        classname = op.splitext(op.basename(filename))[0]
        with open(filename, 'rb') as fp:
            hsh = decodeBinary(fp.read())

        schema = Schema(name=classname, hash=hsh)
        binding = build_binding(schema)

        config_filename = op.splitext(filename)[0] + '.config'
        with open(config_filename, 'rb') as fp:
            config = decodeBinary(fp.read())

        apply_configuration(config, binding)
        extracted = extract_configuration(binding)

        # Check that the configuration was applied
        for key, value, _ in Hash.flat_iterall(config):
            assert key in extracted
            if isinstance(value, np.ndarray):
                assert all(extracted[key] == value)
            else:
                assert extracted[key] == value


def test_vector_hash_extract():
    """Test that we can extract vector hashes from a binding"""

    def get_table_schema(default_value):
        class RowSchema(Configurable):
            value = Int32(defaultValue=0)

        class TableTest(Configurable):
            table = VectorHash(
                defaultValue=default_value,
                rows=RowSchema)
            integer = Int32(defaultValue=1)

        return TableTest.getClassSchema()

    binding = build_binding(get_table_schema(None))
    apply_default_configuration(binding)
    extracted = extract_configuration(binding)
    assert isinstance(extracted, Hash)
    assert "table" not in extracted
    assert "integer" in extracted
    assert extracted["integer"] == 1

    binding = build_binding(get_table_schema([Hash("value", 0)]))
    apply_default_configuration(binding)
    extracted = extract_configuration(binding)
    assert isinstance(extracted, Hash)
    assert "table" in extracted
    assert "integer" in extracted
    assert extracted["integer"] == 1
    assert extracted["table"] == [Hash("value", 0)]


def test_complete_schema():
    schema = get_all_props_schema()
    binding = build_binding(schema)

    namespace = binding.value
    for name in namespace:
        node = getattr(namespace, name)
        assert isinstance(node, ALL_PROPERTIES_MAP[name])


def test_build_binding_stability():
    schema = get_all_props_schema()
    binding = build_binding(schema)

    before_names = list(binding.value)
    binding = build_binding(schema, existing=binding)
    after_names = list(binding.value)

    # The order must be unchanged!
    assert before_names == after_names


def test_default_values():
    schema = get_all_props_schema()
    binding = build_binding(schema)

    assert binding.value.a.value is Undefined
    assert binding.value.b.value is Undefined

    apply_default_configuration(binding)

    assert binding.value.a.value
    assert binding.value.b.value == 'c'

    # Make sure the extracted default conversion is minimal
    # It should include properties with default values, options, or node types
    config = extract_configuration(binding)
    default_props = ('a', 'b', 'h1')
    for prop in default_props:
        assert prop in config, f'{prop!r} missing from config'
    assert "k1" not in config
    assert 'm' not in config
    assert 'i1' not in config, "recursive in config"
    assert 'j1' not in config, "recursive in config"
    # Since no user edits, an empty Hash should have been extracted
    config = extract_edits(schema, binding)
    assert config == Hash()


def test_apply_configuration():
    schema = get_all_props_schema()
    binding = build_binding(schema)

    config = Hash('a', False)
    # Applying a configuration always notifies listeners
    with watch_config_update_notification(binding, expected=True):
        apply_configuration(config, binding)
    assert not binding.value.a.value

    config = Hash('not', 'exist')
    apply_configuration(config, binding)
    # Non exist property is ignored
    assert 'not' not in binding.value

    config = Hash('mm', b'foo')
    # bytes type value is converted to bytearray by traits handler
    apply_configuration(config, binding)
    assert binding.value.mm.value == bytearray(b'foo')


def test_apply_project_configuration():
    schema = get_all_props_schema()
    binding = build_binding(schema)

    config = Hash('a', False)
    # Applying a project configuration always does not notify listeners
    with watch_config_update_notification(binding, expected=False):
        apply_project_configuration(config, binding)
    assert not binding.value.a.value

    config = Hash('not', 'exist')
    apply_project_configuration(config, binding)
    # Non exist property is ignored
    assert 'not' not in binding.value

    config = Hash('mm', b'foo')
    # bytes type value is converted to bytearray by traits handler
    apply_project_configuration(config, binding)
    assert binding.value.mm.value == bytearray(b'foo')

    # try to apply a wrong table value from a project device
    config = Hash('x', HashList([Hash('start', 'not a string', 'stop', 10)]))
    apply_project_configuration(config, binding)
    assert binding.value.x.value == HashList(
        [Hash('start', 0.0, 'stop', 10.0)])


def test_apply_fast_data():
    schema = get_all_props_schema()
    binding = build_binding(schema)
    ts = Timestamp()
    ts.tid = 1337

    config = Hash('a', False)
    with watch_config_update_notification(binding, expected=True):
        apply_fast_data(config, binding, ts)
    assert not binding.value.a.value
    assert binding.value.a.timestamp.tid == 1337

    config = Hash('a', True)
    # configuration is applied but no notification fired
    with watch_config_update_notification(binding, expected=True):
        apply_fast_data(config, binding, ts)
    assert binding.value.a.value
    assert binding.value.a.timestamp.tid == 1337


def test_extract_edit():
    schema = get_all_props_schema()
    binding = build_binding(schema)
    apply_default_configuration(binding)

    # 'a' has default value is True, give 'e' an alarm low attribute
    # XXX: j1 will always be extracted
    config = Hash('a', False, 'e', 0.0)
    apply_project_configuration(config, binding)

    extracted = extract_edits(schema, binding)
    assert extracted == config


def test_extract_online_edit():
    schema = get_all_props_schema()
    binding = build_binding(schema)
    apply_default_configuration(binding)

    # 'a' has default value is True
    config = Hash('a', False, 'e', 0.0, 'j1', [])
    apply_project_configuration(config, binding)

    extracted = extract_online_edits(schema, binding)
    assert extracted == Hash('a', False, 'e', 0.0)

    schema = get_simple_schema()
    binding = build_binding(schema)

    # New round
    apply_default_configuration(binding)
    config = Hash('foo', False, 'bar', 'now', 'charlie', 21)
    apply_project_configuration(config, binding)
    extracted = extract_online_edits(schema, binding)
    # Charlie is not writable
    assert extracted == Hash('foo', False, 'bar', 'now')

    # New round, no changes
    apply_default_configuration(binding)
    extracted = extract_online_edits(schema, binding)
    assert extracted == Hash()

    # New round, change foo
    apply_default_configuration(binding)
    config = Hash('foo', False, 'bar', 'default')
    apply_project_configuration(config, binding)
    extracted = extract_online_edits(schema, binding)
    assert extracted == Hash('foo', False)


def test_extract_online_devicenode():
    class Device(Configurable):
        dn = String(displayType="deviceNode")

    schema = Device.getClassSchema()
    binding = build_binding(schema)
    apply_default_configuration(binding)
    extracted = extract_online_edits(schema, binding)
    assert extracted == Hash()

    # DeviceNodes with a value are successful
    config = Hash('dn', "XHQ_EG_CTRL/MDL/1")
    apply_project_configuration(config, binding)
    extracted = extract_online_edits(schema, binding)
    assert extracted == config


def test_property_attributes():
    schema = get_all_props_schema()
    binding = build_binding(schema)

    assert binding.value.i.accessMode is AccessMode.READONLY
    assert binding.value.i.assignment is Assignment.INTERNAL
    assert binding.value.j.unit_label == 'm'

    assert binding.value.a.accessMode is AccessMode.RECONFIGURABLE
    assert binding.value.a.assignment is Assignment.OPTIONAL
    assert binding.value.a.requiredAccessLevel is AccessLevel.EXPERT

    assert binding.value.m.options == ['foo', 'bar', 'baz', 'qux']

    assert binding.value.k1.is_allowed(State.INTERLOCKED)
    assert binding.value.k1.is_allowed(State.ACTIVE.value)


def test_extract_reconfigurable_configuration():
    schema = get_simple_props_schema()
    binding = build_binding(schema)
    config = Hash(
        "boolProperty", False,
        "falseProperty", False,  # Same as default
        "doubleProperty", 2.1,
        "floatProperty", 1.0,  # ReadOnly, but we apply alarm attr
        "intProperty", 0,  # ReadOnly
        "stringProperty", "",
        "table", [],
        "node.charlie", 27,  # Readonly integer
        "node.foo", False,
        "node.bar", "default",  # Same as default
        "internal", False,  # Other value as default, but Assignment.INTERNAL
    )
    config["floatProperty", ...].update({"alarmLow": 2})

    extracted = extract_init_configuration(binding, config)
    assert "boolProperty" in extracted
    bool_attrs = extracted["boolProperty", ...]
    assert "daqPolicy" not in bool_attrs
    assert "doubleProperty" in extracted
    assert "falseProperty" not in extracted
    assert "floatProperty" not in extracted
    assert "intProperty" not in extracted
    assert "stringProperty" in extracted
    assert "table" in extracted
    assert "node.charlie" not in extracted
    assert "node.foo" in extracted
    foo_attr = extracted["node.foo", ...]
    # attributes belong to schema and are not considered for configuration
    assert not foo_attr
    assert "node.bar" not in extracted
    # Assignment.INTERNAL property not considered!
    assert "internal" not in extracted



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on October 26, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from collections import OrderedDict, namedtuple
from operator import attrgetter

from qtpy import uic
from qtpy.QtCore import (
    QAbstractTableModel, QItemSelection, QModelIndex, QSortFilterProxyModel,
    Qt, Slot)
from qtpy.QtWidgets import (
    QAbstractButton, QButtonGroup, QDialog, QDialogButtonBox)

from karabogui.events import (
    KaraboEvent, register_for_broadcasts, unregister_from_broadcasts)
from karabogui.logger import get_logger
from karabogui.singletons.api import get_config, get_db_conn
from karabogui.util import InputValidator, SignalBlocker, utc_to_local

from .utils import get_dialog_ui

SIMPLE_NAME = "simple_name"
LAST_MODIFIED = "last_modified"
UUID = "uuid"

PROJECT_DATA = OrderedDict()
PROJECT_DATA[SIMPLE_NAME] = "Name"
PROJECT_DATA[LAST_MODIFIED] = "Last Modified"
PROJECT_DATA[UUID] = "UUID"

ProjectEntry = namedtuple("ProjectEntry", list(PROJECT_DATA))
HEADER = list(PROJECT_DATA.values())


def get_column_index(project_data_key):
    """Return `index` in ``PROJECT_DATA`` for `project_data_key`"""
    return list(PROJECT_DATA.keys()).index(project_data_key)


class LoadProjectDialog(QDialog):
    def __init__(self, is_subproject=False, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui("project_handle.ui"), self)
        # set proper window flags
        self.setWindowFlags(self.windowFlags() | Qt.WindowCloseButtonHint)

        db_conn = get_db_conn()
        self.rbFromRemote.setChecked(db_conn.ignore_local_cache)
        self.rbFromCache.setChecked(not db_conn.ignore_local_cache)
        self.load_from_group = QButtonGroup(self)
        self.load_from_group.addButton(self.rbFromRemote)
        self.load_from_group.addButton(self.rbFromCache)
        self.load_from_group.buttonClicked.connect(self._open_from_changed)

        if is_subproject:
            title = "Load Sub Project"
        else:
            title = "Load Master Project"
        self.setWindowTitle(title)
        self.buttonBox.button(QDialogButtonBox.Ok).setText("Load")

        self.model = QSortFilterProxyModel(parent=self)
        # Set up the filter model!
        self.model.setSourceModel(TableModel(parent=self))
        self.model.setFilterRole(Qt.DisplayRole)
        self.model.setFilterFixedString("")
        self.model.setFilterCaseSensitivity(False)
        self.model.setFilterKeyColumn(0)

        column = int(get_config()["project_sort_column"])
        order = int(get_config()["project_sort_order"])
        self.twProjects.horizontalHeader().setSortIndicator(column, order)

        # QTableview in ui file
        self.twProjects.setModel(self.model)
        self.twProjects.selectionModel().selectionChanged.connect(
            self._selectionChanged)
        self.twProjects.doubleClicked.connect(self._load_item)
        self.twProjects.horizontalHeader().sortIndicatorChanged.connect(
            self._sorting_changed)
        # Domain is not selectable for subprojects - only master projects
        self.cbDomain.setEnabled(not is_subproject)
        # ... request the domains list
        # initial request
        self.initial_request = False
        domains = db_conn.get_available_domains()

        # Domain combobox
        if is_subproject:
            default_domain = db_conn.default_domain
        else:
            topic = get_config()["broker_topic"]
            default_domain = (topic if topic in domains
                              else db_conn.default_domain)
        self.default_domain = default_domain

        if not self.ignore_cache:
            # Only fill with the cache domains if the user has requested it!
            self._domains_updated(domains)

        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.leTitle.textChanged.connect(self._titleChanged)
        self.pbClear.clicked.connect(self.leTitle.clear)

        self.event_map = {
            KaraboEvent.ProjectItemsList: self._event_item_list,
            KaraboEvent.ProjectDomainsList: self._event_domain_list,
            KaraboEvent.ProjectTrashed: self._event_trashed
        }
        register_for_broadcasts(self.event_map)

        # Set the focus on the search title!
        self.leTitle.setFocus()

    # -----------------------------------------------------------------------
    # Karabo Events

    def _event_item_list(self, data):
        items = data.get("items", [])
        self.model.sourceModel().add_project_manager_data(items)
        # NOTE: Resize all columns until qtpy
        self.twProjects.resizeColumnsToContents()
        self._titleChanged(self.leTitle.text())

    def _event_domain_list(self, data):
        self._domains_updated(data.get("items", []))

    def _event_trashed(self, data):
        domain = data.get("domain", "")
        self.on_cbDomain_currentIndexChanged(domain)

    # -----------------------------------------------------------------------

    def done(self, result):
        """ Reimplement ``QDialog`` virtual slot

        Stop listening for broadcast events
        """
        unregister_from_broadcasts(self.event_map)
        super().done(result)

    def _domains_updated(self, domains):
        # Domain combobox
        with SignalBlocker(self.cbDomain):
            self.cbDomain.clear()
            self.cbDomain.addItems(sorted(domains))
        # Select default domain, can be initially `None`
        index = 0
        if self.default_domain is not None:
            index = self.cbDomain.findText(self.default_domain)
        with SignalBlocker(self.cbDomain):
            self.cbDomain.setCurrentIndex(index if index > -1 else 0)
        if not self.initial_request:
            self.initial_request = True
            self.update_view()

    def selected_item(self):
        """Return selected domain and project
        """
        selection_model = self.twProjects.selectionModel()
        uuid_index = get_column_index(UUID)
        uuid_entry = selection_model.selectedRows(uuid_index)
        if uuid_entry:
            return self.domain, uuid_entry[0].data()
        return None, None

    def update_view(self):
        if self.domain:
            self.twProjects.clearSelection()
            model = self.model.sourceModel()
            model.request_data(self.domain, self.ignore_cache)

    @property
    def ignore_cache(self):
        return self.rbFromRemote.isChecked()

    @property
    def simple_name(self):
        rows = self.twProjects.selectionModel().selectedRows()
        if rows:
            return rows[0].data()

    @property
    def domain(self):
        return self.cbDomain.currentText()

    def _selected_item_loadable(self):
        """ Return whether the currently selected is project loadable.
        """
        col_index = get_column_index(UUID)
        if self.twProjects.selectionModel().selectedRows(col_index):
            return True

        return False

    def _check_button_state(self):
        # Check if we have a preceeding valid selection
        enable = self._selected_item_loadable()
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enable)

    @Slot(int, Qt.SortOrder)
    def _sorting_changed(self, column, order):
        """Change the view when sorting the table"""
        get_config()["project_sort_column"] = column
        get_config()["project_sort_order"] = order
        self.update_view()

    @Slot(str)
    def _titleChanged(self, text):
        self.model.setFilterFixedString(text)
        self._check_button_state()

    @Slot(QItemSelection, QItemSelection)
    def _selectionChanged(self, selected, deselected):
        """ Whenever an item is selected the current title and the button box
        need to be updated
        """
        # Make sure loading of trashed projects is not possible
        self._check_button_state()

    @Slot(QModelIndex)
    def _load_item(self, index):
        """ Slot connectect to the ``QTableView`` signal ``doubleClicked``
        Only accept the dialog, if the selected project is loadable.
        """
        # Make sure loading of trashed projects is not possible
        if self._selected_item_loadable():
            index = index.sibling(index.row(), 0)
            simple_name = self.model.data(index, Qt.DisplayRole)
            text = (f"Loading project <b>{simple_name}</b> from project "
                    "database")
            get_logger().info(text)
            self.accept()

    @Slot(str)
    def on_cbDomain_currentIndexChanged(self, domain):
        self.update_view()

    @Slot(bool)
    def on_cbShowTrash_toggled(self, is_checked):
        model = self.model.sourceModel()
        model.show_trashed = is_checked

        self._check_button_state()
        self.update_view()

    @Slot(QAbstractButton)
    def _open_from_changed(self, button):
        self.update_view()


class NewProjectDialog(QDialog):
    def __init__(self, model=None, is_rename=False, default=False,
                 parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui("project_new.ui"), self)
        self.setModal(False)

        validator = InputValidator()

        # Domain combobox
        db_conn = get_db_conn()
        domains = db_conn.get_available_domains()
        if default:
            default_domain = db_conn.default_domain
        else:
            topic = get_config()["broker_topic"]
            default_domain = (topic if topic in domains
                              else db_conn.default_domain)
        self.default_domain = default_domain
        self._fill_domain_combo_box(domains)
        # Subprojects reside in the domain of the parent project, only allow
        # the default!
        self.cbDomain.setEnabled(not default)

        if model is None:
            title = "New project"
        else:
            if is_rename:
                title = "Rename project"
                text = model.simple_name
                # Hide domain related widgets
                self.laDomain.hide()
                self.cbDomain.hide()
                self.adjustSize()
            else:
                title = "Create a copy of this project..."
                text = f"{model.simple_name}_copy"
            self.leTitle.setText(text)
        self.setWindowTitle(title)
        self.leTitle.setFocus()
        self.leTitle.selectAll()
        self.leTitle.setValidator(validator)
        self.leTitle.textChanged.connect(self.validate)
        self.validate()

        self.event_map = {
            KaraboEvent.ProjectDomainsList: self._event_item_list
        }
        register_for_broadcasts(self.event_map)

    # -----------------------------------------------------------------------
    # PyQt Slots

    @Slot()
    def validate(self):
        enabled = self.leTitle.hasAcceptableInput()
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)

    # -----------------------------------------------------------------------
    # Karabo Events

    def _event_item_list(self, data):
        self._fill_domain_combo_box(data.get("items", []))

    # -----------------------------------------------------------------------

    def done(self, result):
        """ Reimplement ``QDialog`` virtual slot

        Stop listening for broadcast events
        """
        unregister_from_broadcasts(self.event_map)
        super().done(result)

    @property
    def simple_name(self):
        return self.leTitle.text()

    @property
    def domain(self):
        return self.cbDomain.currentText()

    def _fill_domain_combo_box(self, domains):
        """Put the given list of ``domains`` into the given ``domain_combo``
        """
        with SignalBlocker(self.cbDomain):
            self.cbDomain.clear()
            self.cbDomain.addItems(sorted(domains))
        # Select default domain
        index = self.cbDomain.findText(self.default_domain)
        self.cbDomain.setCurrentIndex(index if index > -1 else 0)


class TableModel(QAbstractTableModel):
    headers = HEADER

    def __init__(self, show_trashed=False, parent=None):
        super().__init__(parent)
        self.entries = []
        self.db_conn = get_db_conn()
        self.show_trashed = show_trashed

    def request_data(self, domain, ignore_cache):
        """Request data for the given ``domain``
        """
        self.db_conn.ignore_local_cache = ignore_cache
        project_data = self.db_conn.get_available_project_data(domain,
                                                               "project")
        self.add_project_manager_data(project_data)

    def add_project_manager_data(self, data):
        """ Add the given `data` to the internal data structure

        :param data: A `HashList` with the keys per entry:
                     - 'uuid' - The unique ID of the Project
                     - 'simple_name' - The name for displaying
                     - 'is_trashed' - Flag if project is marked as trashed
                     - 'date' - The date the project was last modified

        Note: The `user` information is not used.
        """
        # XXX: this only works if the sent list of uuids is complete
        self.beginResetModel()
        try:
            self.entries = []
            for it in data:
                is_trashed = it.get("is_trashed")
                if is_trashed != self.show_trashed:
                    continue
                entry = ProjectEntry(
                    simple_name=it.get("simple_name"),
                    last_modified=utc_to_local(it.get("date")),
                    uuid=(it.get("uuid"), is_trashed))
                self.entries.append(entry)
        finally:
            self.endResetModel()

        # Sort the table according to the column
        column = int(get_config()["project_sort_column"])
        order = int(get_config()["project_sort_order"])
        self.sort(column, order)

    def rowCount(self, parent=None):
        return len(self.entries)

    def columnCount(self, parent=None):
        return len(self.headers)

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None
        entry = self.entries[index.row()]
        column = index.column()
        if role == Qt.DisplayRole:
            if column == get_column_index(UUID):
                # Only display uuid here, ignore is_trashed
                uuid, _ = entry[column]
                return uuid
            return entry[column]
        elif role == Qt.UserRole:
            return entry[column]
        return None

    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.headers[section]

    def sort(self, column, order=Qt.AscendingOrder):
        """ Sort table by given column number and order """
        self.layoutAboutToBeChanged.emit()
        self.entries.sort(key=attrgetter(ProjectEntry._fields[column]),
                          reverse=bool(order))
        self.layoutChanged.emit()
        super().sort(column, order)



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 6, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import time
from contextlib import contextmanager

from qtpy.QtCore import QEvent, QPoint, QRect, QSize, Qt, QTimer, Signal, Slot
from qtpy.QtGui import QBrush, QColor, QPainter, QPalette, QPen
from qtpy.QtWidgets import QSizePolicy, QStackedLayout, QWidget

import karabogui.access as krb_access
from karabo.common.api import set_initialized_flag
from karabo.common.scenemodel.api import (
    SCENE_MIN_HEIGHT, SCENE_MIN_WIDTH, FixedLayoutModel)
from karabogui.binding.api import DeviceProxy
from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts,
    unregister_from_broadcasts)
from karabogui.generic_scenes import get_property_proxy_model
from karabogui.request import retrieve_default_scene, send_property_changes

from .bases import BaseSceneTool
from .builder import (
    bring_object_to_front, create_object_from_model, fill_root_layout,
    find_top_level_model, is_rect, is_widget, iter_widgets_and_models,
    remove_object_from_layout, replace_model_in_top_level_model,
    send_object_to_back)
from .const import QT_CURSORS, SELECTION_COLOR
from .layout.api import GroupLayout
from .selection_model import SceneSelectionModel
from .tools.api import (
    ConfigurationDropHandler, NavigationDropHandler, ProjectDropHandler,
    ProxySelectionTool, SceneControllerHandler, SceneSelectionTool,
    SceneToolHandler)
from .utils import save_painter_state
from .widget.api import ControllerContainer, GridView

# The scene widgets handler for mutations and action of the controllers and
# layout items on our scene
SCENE_WIDGET_HANDLER = (SceneControllerHandler, SceneToolHandler)
WIDGET_HANDLES_SIZE = QSize(5, 5)
WIDGET_HANDLES_OFFSET = QPoint(int(WIDGET_HANDLES_SIZE.width() / 2),
                               int(WIDGET_HANDLES_SIZE.height() / 2))

_WIDGET_REMOVAL_DELAY = 5000
_VISIBILITY_TIMEOUT = 3000
_REMOVAL_TIMEOUT = 1000

SELECTION_QCOLOR = QColor(*SELECTION_COLOR)


def _get_time_milli():
    return int(round(time.time() * 1000))


class SceneView(QWidget):
    """An object representing the view for a Karabo GUI scene.
    """

    resetToSelectionTool = Signal()

    def __init__(self, model=None, parent=None):
        super().__init__(parent)

        layout_model = FixedLayoutModel(x=0, y=0, width=SCENE_MIN_WIDTH,
                                        height=SCENE_MIN_HEIGHT)
        # XXX: This is an interesting hack.
        # Basically, we want to put all of the scene's children into a widget
        # which is parented to this widget and completely covers it.
        # Then, the `design_mode` property toggles the mouse handling for this
        # child widget. In design mode, we want the main scene view and its
        # tools to get the mouse events. Out of design mode, we re-enable mouse
        # handling for all of the scene's children.
        self.layout = GroupLayout(layout_model, parent=self)
        self.setContentsMargins(0, 0, 0, 0)

        self.inner = GridView(self)
        self.inner.setLayout(self.layout)
        self.inner.setFocusPolicy(Qt.StrongFocus)

        layout = QStackedLayout(self)
        layout.setStackingMode(QStackedLayout.StackAll)
        layout.addWidget(self.inner)

        self.scene_model = None
        self._ignore_resize_events = False
        self.selection_model = SceneSelectionModel()

        # List of scene drag n drop handlers
        self.scene_handler_list = [ConfigurationDropHandler(),
                                   NavigationDropHandler(),
                                   ProjectDropHandler()]
        self.current_scene_handler = None

        self.current_tool = None
        # Use an internal flag to explicitly mark the design_mode, so when
        # other tools need the parent widget to handle mouse event won't
        # trigger the design_mode behavior
        self._design_mode = False
        self.snap_to_grid = True
        self.set_scene_edit_mode(False)
        self.tab_visible = False
        self._scene_obj_cache = {}
        self._hover_rect = QRect(0, 0, 10, 10)

        # Widget cleanup
        self._widget_removal_queue = []
        self._widget_removal_timer = QTimer(self)
        self._widget_removal_timer.setInterval(_REMOVAL_TIMEOUT)
        self._widget_removal_timer.timeout.connect(self._clean_removed_widgets)

        self._visible_timer = QTimer(self)
        self._visible_timer.setInterval(_VISIBILITY_TIMEOUT)
        self._visible_timer.setSingleShot(True)
        self._visible_timer.timeout.connect(self._clean_visible_widgets)

        self.setFocusPolicy(Qt.StrongFocus)
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.setAcceptDrops(True)
        self.setAttribute(Qt.WA_MouseTracking)
        self.setBackgroundRole(QPalette.Window)
        self.resize(SCENE_MIN_WIDTH, SCENE_MIN_HEIGHT)

        self.update_model(model)

        self.event_map = {
            KaraboEvent.AccessLevelChanged: self._event_access_level
        }
        register_for_broadcasts(self.event_map)

    def set_scene_edit_mode(self, value):
        """Set the edit mode of the scene view and toggle mouse handling in
        the inner view"""
        self.enable_mouse_event_handling(value)
        if not value:
            self.selection_model.clear_selection()
            self.set_tool(None)

    @property
    def design_mode(self):
        return self._design_mode

    @design_mode.setter
    def design_mode(self, value):
        self._design_mode = value
        self.set_scene_edit_mode(value)
        self.inner.showGrid(value)

    def enable_mouse_event_handling(self, enable):
        """Set whether delivery of mouse events to the widget and its children
        is either enabled or disabled.
        """
        self.inner.setAttribute(Qt.WA_TransparentForMouseEvents, enable)

    @contextmanager
    def ignore_resize_events(self):
        """Provide a context for explicitly disabling resize propagation to
        the data model.

        This is necessary because Qt will first set us to a size which it
        chooses when undocking, then give us a chance to set the size we
        actually want. In the meantime, it would be nice to disable updates
        so that our scene doesn't get modified.
        """
        try:
            self._ignore_resize_events = True
            yield
        finally:
            self._ignore_resize_events = False

    # ----------------------------
    # Qt Methods

    def mouseMoveEvent(self, event):
        if self.design_mode:
            self.current_tool.mouse_move(self, event)
            if event.isAccepted():
                self.update()
            else:
                super().mouseMoveEvent(event)

    def mousePressEvent(self, event):
        proxy_selecting = isinstance(self.current_tool, ProxySelectionTool)
        if self.design_mode or proxy_selecting:
            self.current_tool.mouse_down(self, event)
            if event.isAccepted():
                self.update()
            else:
                super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        if self.design_mode:
            self.current_tool.mouse_up(self, event)
            if event.isAccepted():
                self.update()
            else:
                super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event):
        if self.design_mode:
            item = self.item_at_position(event.pos())
            if item is not None and hasattr(item, 'edit'):
                item.edit(self)
                self.update()
            event.accept()
        else:
            item = self.controller_at_position(event.pos())
            if item is None or item.is_editable:
                event.ignore()
                return
            proxy_selecting = isinstance(self.current_tool, ProxySelectionTool)
            proxy = item.widget_controller.proxy
            if proxy_selecting:
                device = None
                if (isinstance(proxy.root_proxy, DeviceProxy)
                        and proxy.root_proxy.online):
                    device = proxy.root_proxy
                broadcast_event(KaraboEvent.RaiseEditor, {"proxy": device})
            else:
                # If we having a control modifier, we ask for the scene
                if event.modifiers() & Qt.ControlModifier:
                    retrieve_default_scene(proxy.root_proxy.device_id)
                    return
                # Get a controller widget
                model = get_property_proxy_model(proxy, include_images=False)
                if model is not None:
                    data = {"model": model}
                    broadcast_event(KaraboEvent.ShowUnattachedController, data)

            event.accept()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Alt and not self.design_mode:
            if not self.inner.hasFocus():
                self.inner.setFocus()
            self.set_cursor('pointing-hand')
            self.set_tool(ProxySelectionTool())
            self.enable_mouse_event_handling(True)

    def keyReleaseEvent(self, event):
        if event.key() == Qt.Key_Alt and not self.design_mode:
            self.set_cursor('none')
            self.enable_mouse_event_handling(False)
            self.set_tool(None)

    def dragEnterEvent(self, event):
        if not self.design_mode:
            event.ignore()
            return

        for scene_handler in self.scene_handler_list:
            if scene_handler.can_handle(event):
                self.current_scene_handler = scene_handler
                event.accept()
                return
        event.ignore()

    def dropEvent(self, event):
        if self.current_scene_handler is None:
            event.ignore()
            return

        self.current_scene_handler.handle(self, event)
        self.current_scene_handler = None
        event.accept()

    def paintEvent(self, event):
        """Show view content.
        """
        with QPainter(self) as painter:
            self.layout.draw(painter)
            self._draw_selection(painter)

            current_tool = self.current_tool
            if current_tool.visible:
                current_tool.draw(self, painter)

    def sizeHint(self):
        model = self.scene_model
        if model is None:
            return QSize()

        return QSize(model.width, model.height)

    def resizeEvent(self, event):
        model = self.scene_model
        if model is None:
            return

        if self._ignore_resize_events or not self.design_mode:
            # Often, we want to avoid modifying the model
            return

        size = event.size()
        model.width = size.width()
        model.height = size.height()

    def event(self, event):
        """This needs to be reimplemented to show tooltips not only in control
        mode but also in design mode.

        If the design mode is active the widget attribute
        ``WA_TransparentForMouseEvents`` is set to ``False`` which prevents the
        forwarding of the events to the actual widgets.
        """
        if event.type() == QEvent.ToolTip:
            widget = self.widget_at_position(event.pos())
            if widget is not None:
                return widget.event(event)
        return super().event(event)

    def _event_access_level(self, data):
        self._update_widget_states()

    def contextMenuEvent(self, event):
        """Show scene view specific context menu. """
        if self.design_mode:
            widget = self.widget_at_position(event.pos())
            if widget is None:
                event.ignore()
                return
            # NOTE: An eventual handler blocks here until an action is selected
            for handler in SCENE_WIDGET_HANDLER:
                widget_handler = handler(widget=widget)
                if widget_handler.can_handle():
                    widget_handler.handle(self, event)
                    event.accept()
                    return

        event.ignore()

    # ----------------------------
    # Public methods

    def apply_editor_changes(self):
        """Apply user edits to a remote device instance
        """
        proxies = [proxy for proxy in self.cached_proxies()
                   if proxy.edit_value is not None]
        send_property_changes(proxies)

    def decline_editor_changes(self):
        """Decline user edits in editor widgets
        """
        for obj in self._scene_obj_cache.values():
            if is_widget(obj):
                obj.decline_changes()

    def destroy(self):
        """Do some cleanup of the scene's objects before death.
        """
        self._widget_removal_timer.stop()
        for obj in self._scene_obj_cache.values():
            if is_widget(obj):
                obj.destroy()

        unregister_from_broadcasts(self.event_map)
        self._set_scene_model(None)
        self._scene_obj_cache.clear()
        self._widget_removal_queue.clear()
        super().destroy()

    def set_tab_visible(self, visible):
        """Sets whether this scene is visible.

        This method manages the visibilities of the device properties in
        this scene.
        """
        if not visible:
            self._visible_timer.start()
            return

        # In case of a previous visible trigger we must deactivate the timer!
        if self._visible_timer.isActive():
            self._visible_timer.stop()

        # If we are already visible, we simply return!
        if self.tab_visible:
            return

        for obj in self._scene_obj_cache.values():
            # NOTE: check also if widget is visible due to the fact that we
            # only hide widgets with got replaced by 'Change Widget...'
            if is_widget(obj) and obj.isVisible():
                obj.set_visible(True)
        self.tab_visible = True

    @Slot()
    def _clean_visible_widgets(self):
        """Remove the visibility state of this scene"""
        for obj in self._scene_obj_cache.values():
            # NOTE: We must set the visibility of all widgets to `False`!
            if is_widget(obj):
                obj.set_visible(False)
        self.tab_visible = False

    def update_model(self, scene_model):
        if scene_model is None:
            return

        # Set scene model
        self._set_scene_model(scene_model)
        # Set width and height
        self.resize(self.scene_model.width, self.scene_model.height)

        self._scene_obj_cache = {}
        fill_root_layout(self.layout, self.scene_model, self.inner,
                         self._scene_obj_cache, self.tab_visible)

    def item_at_hover(self, pos):
        """Returns the topmost object whose bounds contain `pos`."""
        self._hover_rect.moveCenter(pos)

        # Prioritize items on the selection
        child = self.selection_model.child_in_rect(self._hover_rect)
        if child is not None and not is_rect(child):
            return child

        for child in self.scene_model.children[::-1]:
            obj = self._scene_obj_cache.get(child)
            if obj is not None and obj.geometry().intersects(self._hover_rect):
                return obj

    def item_at_position(self, pos):
        """Returns the topmost object whose bounds contain `pos`."""

        # Prioritize items on the selection
        child = self.selection_model.child_in_pos(pos)
        if child is not None:
            return child

        for child in self.scene_model.children[::-1]:
            obj = self._scene_obj_cache.get(child)
            if obj is not None and obj.geometry().contains(pos):
                return obj

    def controller_at_position(self, pos):
        """Returns the topmost controller whose bounds contain `pos`."""
        widget = self.inner.childAt(pos)
        while (widget is not None
               and not isinstance(widget, ControllerContainer)):
            widget = widget.parent()

        return widget

    def widget_at_position(self, pos):
        """Returns the topmost widget whose bounds contain `pos`."""
        widget = self.inner.childAt(pos)
        while (widget is not None
               and not is_widget(widget)):
            widget = widget.parent()
        return widget

    def items_in_rect(self, rect):
        """Returns the objects whose bounds are contained in `rect`.
           This respects the object order (back-to-front)"""
        items = []
        for child in self.scene_model.children:
            obj = self._scene_obj_cache.get(child)
            if obj is not None and rect.contains(obj.geometry()):
                items.append(obj)
        return items

    def select_all(self):
        """Add all children to the current selection"""
        self.selection_model.clear_selection()
        for child in self.scene_model.children:
            obj = self._scene_obj_cache.get(child)
            if obj is not None:
                self.selection_model.select_object(obj)
        self.update()

    def select_model(self, model):
        """Select widget object for the given ``model``."""
        self.selection_model.clear_selection()
        obj = self._scene_obj_cache.get(model)
        if obj is not None:
            self.selection_model.select_object(obj)
        self.update()

    def select_models(self, *models):
        """Select multiple widget objects for the given ``models``.
        This is used by the scene paste action"""
        self.selection_model.clear_selection()
        for model in models:
            obj = self._scene_obj_cache.get(model)
            if obj is not None:
                self.selection_model.select_object(obj)
        self.update()

    def set_cursor(self, name):
        """Sets the cursor for the scene view."""
        if name == 'none':
            self.unsetCursor()
        else:
            self.setCursor(QT_CURSORS[name])

    def set_tool(self, tool):
        """Sets the current tool being used by the view."""
        assert tool is None or isinstance(tool, BaseSceneTool)
        self.current_tool = tool
        if tool is None:
            if self.design_mode:
                self.resetToSelectionTool.emit()
            else:
                self.set_cursor('none')
                self.current_tool = SceneSelectionTool()

        self.update()

    def add_models(self, *models, initialize=False):
        """Adds new child models to the scene model."""
        if initialize:
            for model in models:
                set_initialized_flag(model, value=True)

        self.scene_model.children.extend(models)

    def remove_model(self, model):
        """Removes the given ``model`` from the scene model."""
        self.scene_model.children.remove(model)

    def replace_model(self, old_model, new_model):
        """Replace the given ``old_model`` with the given ``new_model``."""
        # Find top level model to which ``old_model`` belongs
        top_level_model = find_top_level_model(self.scene_model, old_model)
        if top_level_model is None:
            return

        # Remove it from list
        self.remove_model(top_level_model)
        if top_level_model is not old_model:
            # Do the replacing in the top level model tree
            replace_model_in_top_level_model(self.scene_model,
                                             top_level_model,
                                             old_model, new_model)
        else:
            top_level_model = new_model
        # Add the modified top level model to the scene again
        self.add_models(top_level_model)
        self.selection_model.clear_selection()

    def bring_to_front(self, model):
        """The given ``model`` is moved to the end of the list."""
        # Remove model
        self.remove_model(model)
        # Add model to end
        self.add_models(model)

        scene_obj = self._scene_obj_cache.get(model)
        if scene_obj is not None:
            # In case of layouts or widgets
            bring_object_to_front(scene_obj)

    def send_to_back(self, model):
        """The given ``model`` is moved to the beginning of the list."""
        # Copy list to trigger traits notification only once when its set at
        # the end
        children = list(self.scene_model.children)
        # Clean list to redraw all models in correct order
        del self.scene_model.children[:]
        # Remove model
        children.remove(model)
        # Insert model at beginning
        children.insert(0, model)
        self.scene_model.children.extend(children)

        scene_obj = self._scene_obj_cache.get(model)
        if scene_obj is not None:
            # In case of layouts or widgets
            send_object_to_back(scene_obj)

    def mouse_left(self):
        """mouse left scene view area"""
        if not self.design_mode:
            # Deactivate special tools if not in design mode
            self.set_cursor('none')
            self.set_tool(None)
            self.enable_mouse_event_handling(False)

    def cached_proxies(self):
        """Return all available property proxies from the scene obj cache"""
        widgets = [obj for obj in self._scene_obj_cache.values()
                   if isinstance(obj, ControllerContainer)]
        proxies = [proxy for widget in widgets
                   for proxy in widget.widget_controller.proxies]
        return proxies

    # --------------------------------------------------------------------
    # Private methods

    @Slot()
    def _clean_removed_widgets(self):
        """Called by the `_widget_removal_timer` when there might be widgets to
        clean up.
        """
        remainders = []
        current_time = _get_time_milli()
        while len(self._widget_removal_queue) > 0:
            removal_time, widget, model = self._widget_removal_queue.pop()
            if widget.isVisible():
                continue  # already re-added. ignore

            if (current_time - removal_time) > _WIDGET_REMOVAL_DELAY:
                # Time's up!
                widget.destroy()
                widget.setParent(None)
                self._scene_obj_cache.pop(model, None)
            else:
                remainders.append((removal_time, widget, model))

        self._widget_removal_queue.extend(remainders)
        if len(self._widget_removal_queue) == 0:
            self._widget_removal_timer.stop()

    def _collect_removed_widgets(self, obj, model):
        """Collect widgets for the `widget_removal_queue`.
        """
        for item in iter_widgets_and_models(obj, model, self._scene_obj_cache):
            # Schedule the (possible) eventual cleanup of the widget
            removal_item = (_get_time_milli(),) + item
            self._widget_removal_queue.append(removal_item)

        if (self._widget_removal_queue and
                not self._widget_removal_timer.isActive()):
            self._widget_removal_timer.start()

    def _draw_selection(self, painter):
        """Draw a dashed rect around the selected objects."""
        selection_model = self.selection_model
        if not selection_model.has_selection():
            return

        # Paint main selection bounds
        rect = selection_model.get_selection_bounds()
        resizable = selection_model.has_resizable_selection()
        with save_painter_state(painter):
            paint_bounds(painter, rect, active_handles=resizable)

    def _set_scene_model(self, scene_model):
        """The scene model is set and all notification handlers are defined.
        """
        if self.scene_model is not None:
            self.scene_model.on_trait_change(self._model_modified,
                                             'children_items', remove=True)

        self.scene_model = scene_model
        if scene_model is None:
            return

        self.scene_model.on_trait_change(self._model_modified,
                                         'children_items')

    def _model_modified(self, event):
        """The scene model got modified."""
        for model in event.removed:
            obj = self._scene_obj_cache.get(model)
            if obj is not None:
                # Collect widgets _before_ removing
                self._collect_removed_widgets(obj, model)
                # Then remove
                remove_object_from_layout(obj, self.layout,
                                          self._scene_obj_cache)

        for model in event.added:
            create_object_from_model(self.layout, model, self.inner,
                                     self._scene_obj_cache, self.tab_visible)

    def _update_widget_states(self):
        """The global access level has changed. Notify all widgets in the
        scene.
        """
        level = krb_access.GLOBAL_ACCESS_LEVEL
        for obj in self._scene_obj_cache.values():
            if is_widget(obj):
                obj.update_global_access_level(level)


def paint_bounds(painter, rect, active_handles=True):
    # Draw borders
    pen = QPen(SELECTION_QCOLOR)
    pen.setStyle(Qt.DashLine)
    painter.setPen(pen)
    painter.setBrush(QBrush(Qt.NoBrush))
    painter.drawRect(rect)

    # Draw squares
    color = SELECTION_QCOLOR if active_handles else Qt.lightGray
    painter.setPen(SELECTION_QCOLOR)
    painter.setBrush(QBrush(color))

    left, top, right, bottom = rect.getCoords()
    center = rect.center()
    x_center, y_center = center.x(), center.y()

    corners = [(x, y) for x in (left, right) for y in (top, bottom)]
    mids = [(left, y_center), (right, y_center), (x_center, top),
            (x_center, bottom)]

    for pos in corners + mids:
        painter.drawRect(QRect(QPoint(*pos) - WIDGET_HANDLES_OFFSET,
                               WIDGET_HANDLES_SIZE))
